"""
Utility functions
"""

import re


def camelcase_to_underscore(camelcase: str) -> str:
    """
    Convert a name in CamelCase to non_camel_case

    :param camelcase: CamelCased string
    :return: non_camel_cased string
    """
    # Handle weird Camel-Case notation
    s0 = camelcase.replace('-', '_')

    # Insert an underscore before any uppercase letter which is followed by a lowercase letter
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s0)

    # Insert an underscore before any uppercase letter which is preceded by a lowercase letter or number
    s2 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)

    # Lowercase the result
    s3 = s2.lower()

    # And return with double underscores collapsed
    return re.sub(r'_+', '_', s3)


def camelcase_to_dash(camelcase: str) -> str:
    """
    Convert a name in CamelCase to non-camel-case

    :param camelcase: CamelCased string
    :return: non-camel-cased string
    """
    # The same as camelcase_to_underscore, but with the underscores replaced by dashes
    return camelcase_to_underscore(camelcase).replace('_', '-')


# Representation and Use of Domain Names
# :rfc:`3315#section-8`
#
# So that domain names may be encoded uniformly, a domain name or a
# list of domain names is encoded using the technique described in
# section 3.1 of :rfc:`1035` [10].  A domain name, or list of domain
# names, in DHCP MUST NOT be stored in compressed form, as described in
# section 4.1.4 of :rfc:`1035`.
def parse_domain_bytes(buffer: bytes, offset: int=0, length: int=None, allow_relative: bool=False) -> (int, str):
    """
    Extract a single domain name.

    :param buffer: The buffer to read data from
    :param offset: The offset in the buffer where to start reading
    :param length: The amount of data we are allowed to read from the buffer
    :param allow_relative: Allow domain names that do not end with a zero-length label
    :return: The number of bytes used from the buffer and the extracted domain name
    """
    my_offset = 0
    max_offset = length or (len(buffer) - offset)

    current_labels = []
    while max_offset > my_offset:
        label_length = buffer[offset + my_offset]
        my_offset += 1

        # End of a sequence of labels
        if label_length == 0:
            domain_name = '.'.join(current_labels)
            return my_offset, domain_name

        if label_length > 63:
            raise ValueError('Domain List contains label with invalid length')

        # Check if we stay below the max offset
        if my_offset + label_length > max_offset:
            raise ValueError('Invalid encoded domain name, exceeds available buffer')

        # New label
        current_label_bytes = buffer[offset + my_offset:offset + my_offset + label_length]
        my_offset += label_length

        current_label = current_label_bytes.decode('ascii')
        current_labels.append(current_label)

    if allow_relative:
        # We have reached the end of the data and we allow relative labels: we're done
        domain_name = '.'.join(current_labels)
        return my_offset, domain_name

    raise ValueError('Domain name must end with a 0-length label')


def parse_domain_list_bytes(buffer: bytes, offset: int=0, length: int=None) -> (int, list):
    """
    Extract a list of domain names.

    :param buffer: The buffer to read data from
    :param offset: The offset in the buffer where to start reading
    :param length: The amount of data we are allowed to read from the buffer
    :return: The number of bytes used from the buffer and the extracted domain names
    """
    my_offset = 0
    max_offset = length or (len(buffer) - offset)

    domain_names = []
    while max_offset > my_offset:
        domain_name_len, domain_name = parse_domain_bytes(buffer,
                                                          offset=offset + my_offset, length=max_offset - my_offset)
        domain_names.append(domain_name)
        my_offset += domain_name_len

    return my_offset, domain_names


def encode_domain(domain_name: str, allow_relative: bool=False) -> bytes:
    """
    Encode a single domain name as a sequence of bytes

    :param domain_name: The domain name
    :param allow_relative: Assume that domain names that don't end with a period are relative and encode them as such
    :return: The encoded domain name as bytes
    """
    buffer = bytearray()

    # Be nice: strip trailing dots
    if allow_relative:
        if domain_name.endswith('.'):
            # Treat as FQDN
            domain_name = domain_name.rstrip('.')
            end_with_zero = True
        else:
            # Treat as relative
            end_with_zero = False
    else:
        # Treat as fqdn
        domain_name = domain_name.rstrip('.')
        end_with_zero = True

    domain_name_parts = domain_name.split('.')
    for label in domain_name_parts:
        label_length = len(label)
        if label_length < 1 or label_length > 63:
            raise ValueError('Domain name contains label with invalid length')

        buffer.append(label_length)
        buffer.extend(label.encode('ascii'))

    if end_with_zero:
        # End FQDN domain name with a 0-length label
        buffer.append(0)

    return buffer


def encode_domain_list(domain_names: [str]) -> bytes:
    """
    Encode a list of domain names to a sequence of bytes

    :param domain_names: The list of domain names
    :return: The encoded domain names as bytes
    """
    buffer = bytearray()
    for domain_name in domain_names:
        buffer.extend(encode_domain(domain_name))
    return buffer



"""
Test the RequestMessage implementation
"""
import unittest
import codecs
from ipaddress import IPv6Network, IPv6Address

from dhcpkit.ipv6.duids import LinkLayerDUID, LinkLayerTimeDUID
from dhcpkit.ipv6.messages import RequestMessage
from dhcpkit.ipv6.options import ElapsedTimeOption, ClientIdOption, IANAOption, \
    ReconfigureAcceptOption, OptionRequestOption, OPTION_IA_NA, OPTION_VENDOR_OPTS, VendorClassOption, \
    IAAddressOption, ServerIdOption
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption, OPTION_IA_PD
from dhcpkit.ipv6.extensions.dns import OPTION_DNS_SERVERS
from dhcpkit.ipv6.extensions.sntp import OPTION_SNTP_SERVERS
from dhcpkit.ipv6.extensions.ntp import OPTION_NTP_SERVER
from dhcpkit.ipv6.extensions.sol_max_rt import OPTION_SOL_MAX_RT, OPTION_INF_MAX_RT
from tests.ipv6.messages import test_client_server_message

""

# DHCPv6
#     Message type: Request (3)
#     Transaction ID: 0xf350d6
#     Elapsed time
#         Option: Elapsed time (8)
#         Length: 2
#         Value: 0068
#         Elapsed time: 104 ms
#     Client Identifier
#         Option: Client Identifier (1)
#         Length: 10
#         Value: 000300013431c43cb2f1
#         DUID: 000300013431c43cb2f1
#         DUID Type: link-layer address (3)
#         Hardware type: Ethernet (1)
#         Link-layer address: 34:31:c4:3c:b2:f1
#     Server Identifier
#         Option: Server Identifier (2)
#         Length: 14
#         Value: 000100011d1d49cf00137265ca42
#         DUID: 000100011d1d49cf00137265ca42
#         DUID Type: link-layer address plus time (1)
#         Hardware type: Ethernet (1)
#         DUID Time: Jun 24, 2015 12:58:23.000000000 CEST
#         Link-layer address: 00:13:72:65:ca:42
#     Identity Association for Non-temporary Address
#         Option: Identity Association for Non-temporary Address (3)
#         Length: 40
#         Value: c43cb2f100000000000000000005001820010db8ffff0001...
#         IAID: c43cb2f1
#         T1: 0
#         T2: 0
#         IA Address
#             Option: IA Address (5)
#             Length: 24
#             Value: 20010db8ffff0001000c00000000e09c0000017700000258
#             IPv6 address: 2001:db8:ffff:1:c::e09c (2001:db8:ffff:1:c::e09c)
#             Preferred lifetime: 375
#             Valid lifetime: 600
#     Identity Association for Prefix Delegation
#         Option: Identity Association for Prefix Delegation (25)
#         Length: 41
#         Value: c43cb2f10000000000000000001a00190000017700000258...
#         IAID: c43cb2f1
#         T1: 0
#         T2: 0
#         IA Prefix
#             Option: IA Prefix (26)
#             Length: 25
#             Value: 00000177000002583820010db8ffccfe0000000000000000...
#             Preferred lifetime: 375
#             Valid lifetime: 600
#             Prefix length: 56
#             Prefix address: 2001:db8:ffcc:fe00:: (2001:db8:ffcc:fe00::)
#     Reconfigure Accept
#         Option: Reconfigure Accept (20)
#         Length: 0
#     Option Request
#         Option: Option Request (6)
#         Length: 16
#         Value: 00170038001f00190003001100520053
#         Requested Option code: DNS recursive name server (23)
#         Requested Option code: NTP Server (56)
#         Requested Option code: Simple Network Time Protocol Server (31)
#         Requested Option code: Identity Association for Prefix Delegation (25)
#         Requested Option code: Identity Association for Non-temporary Address (3)
#         Requested Option code: Vendor-specific Information (17)
#         Requested Option code: SOL_MAX_RT (82)
#         Requested Option code: INF_MAX_RT (83)
#     Vendor Class
#         Option: Vendor Class (16)
#         Length: 4
#         Value: 00000368
#         Enterprise ID: AVM GmbH (872)

request_message = RequestMessage(
    transaction_id=bytes.fromhex('f350d6'),
    options=[
        ElapsedTimeOption(elapsed_time=104),
        ClientIdOption(duid=LinkLayerDUID(hardware_type=1, link_layer_address=bytes.fromhex('3431c43cb2f1'))),
        ServerIdOption(duid=LinkLayerTimeDUID(hardware_type=1, time=488458703,
                                              link_layer_address=bytes.fromhex('00137265ca42'))),
        IANAOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAAddressOption(address=IPv6Address('2001:db8:ffff:1:c::e09c'), preferred_lifetime=375, valid_lifetime=600),
        ]),
        IAPDOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAPrefixOption(prefix=IPv6Network('2001:db8:ffcc:fe00::/56'), preferred_lifetime=375, valid_lifetime=600),
        ]),
        ReconfigureAcceptOption(),
        OptionRequestOption(requested_options=[
            OPTION_DNS_SERVERS,
            OPTION_NTP_SERVER,
            OPTION_SNTP_SERVERS,
            OPTION_IA_PD,
            OPTION_IA_NA,
            OPTION_VENDOR_OPTS,
            OPTION_SOL_MAX_RT,
            OPTION_INF_MAX_RT,
        ]),
        VendorClassOption(enterprise_number=872),
    ],
)

request_packet = codecs.decode('03f350d60008000200680001000a0003'
                               '00013431c43cb2f10002000e00010001'
                               '1d1d49cf00137265ca4200030028c43c'
                               'b2f10000000000000000000500182001'
                               '0db8ffff0001000c00000000e09c0000'
                               '01770000025800190029c43cb2f10000'
                               '000000000000001a0019000001770000'
                               '02583820010db8ffccfe000000000000'
                               '00000000140000000600100017003800'
                               '1f001900030011005200530010000400'
                               '000368', 'hex')


class RequestMessageTestCase(test_client_server_message.ClientServerMessageTestCase):
    def setUp(self):
        self.packet_fixture = request_packet
        self.message_fixture = request_message
        self.parse_packet()


if __name__ == '__main__':
    unittest.main()



"""
Test the UnknownOption implementation
"""
import unittest

from dhcpkit.ipv6.options import UnknownOption
from tests.ipv6.options import test_option


class UnknownOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = b'\x00\xff\x00\x100123456789abcdef'
        self.option_object = UnknownOption(255, b'0123456789abcdef')
        self.parse_option()

    def test_validate_type(self):
        self.option.option_type = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

    def test_validate_data(self):
        self.option.option_data = '0123456789abcdef'
        with self.assertRaisesRegex(ValueError, 'must be sequence of bytes'):
            self.option.validate()

        self.option.option_data = b'0123456789abcdef' * 10000
        with self.assertRaisesRegex(ValueError, 'cannot be longer than'):
            self.option.validate()

    def test_validate_option_type(self):
        # This should be ok
        self.option.option_type = 0
        self.option.validate()

        # This shouldn't
        self.option.option_type = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

        # This should be ok
        self.option.option_type = 65535
        self.option.validate()

        # This shouldn't
        self.option.option_type = 65536
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()


if __name__ == '__main__':
    unittest.main()



"""
DHCP Request handler that just prints incoming requests
"""

import logging

from dhcpkit.ipv6.message_handlers import MessageHandler
from dhcpkit.ipv6.messages import Message, RelayServerMessage

logger = logging.getLogger(__name__)


class DumpRequestsMessageHandler(MessageHandler):
    """
    DHCP Request handler that just prints incoming requests
    """

    def handle(self, received_message: RelayServerMessage, received_over_multicast: bool) -> Message or None:
        """
        Display the received message.

        :param received_message: The parsed incoming request, wrapped in an 'internal' RelayServerMessage
        :param received_over_multicast: Whether the request was received over multicast
        :returns: The message to reply with
        """
        # Print the incoming request
        received_using = 'multicast' if received_over_multicast else 'unicast'
        logger.debug("Received {} message {}".format(received_using, received_message))

        # Not sending any response
        return None


handler = DumpRequestsMessageHandler



"""
Test the basic option implementation
"""
import unittest

from dhcpkit.ipv6.options import UnknownOption, Option


class OptionTestCase(unittest.TestCase):
    def setUp(self):
        # The following attributes must be overruled by child classes
        # The basics are tested with a simple UnknownOption
        self.option_bytes = bytes.fromhex('ffff0010') + b'0123456789abcdef'
        self.option_object = UnknownOption(65535, b'0123456789abcdef')
        self.parse_option()

    def parse_option(self):
        self.length, self.option = Option.parse(self.option_bytes)
        self.assertIsInstance(self.option, Option)
        self.option_class = type(self.option)

    def test_length(self):
        self.assertEqual(self.length, len(self.option_bytes))

    def test_parse(self):
        self.assertEqual(self.option, self.option_object)

    def test_save_parsed(self):
        self.assertEqual(self.option_bytes, self.option.save())

    def test_save_fixture(self):
        self.assertEqual(self.option_bytes, self.option_object.save())

    def test_validate(self):
        # This should be ok
        self.option.validate()

    def test_overflow(self):
        with self.assertRaisesRegex(ValueError, 'longer than .* buffer'):
            self.option_class.parse(self.option_bytes, length=len(self.option_bytes) - 1)

    def test_load_from_wrong_buffer(self):
        if issubclass(self.option_class, UnknownOption):
            # UnknownOption accepts any parsable buffer, no point in testing that one
            return

        option = self.option_class()
        with self.assertRaisesRegex(ValueError, 'buffer does not contain'):
            option.load_from(bytes.fromhex('fffe0000'))


if __name__ == '__main__':
    unittest.main()



"""
Utility functions for option handlers
"""
from collections import namedtuple


# A simple structure to store an address and/or prefix assignment in
from dhcpkit.ipv6.options import Option, StatusCodeOption

Assignment = namedtuple('Assignment', ['address', 'prefix'])


def force_status(options: [Option], new_status_code: StatusCodeOption):
    """
    If there is a StatusCodeOption with a different status code in the options list then replace it. Leave any option
    with the right status code. Add the given StatusCodeOption if there is none.

    :param options: The list of options to manipulate
    :param new_status_code: The wanted StatusCodeOption
    """

    # Check for any existing status options in the response
    existing = [option for option in options if isinstance(option, StatusCodeOption)]
    existing = existing[0] if existing else None
    if existing:
        if existing.status_code == new_status_code.status_code:
            # Ok, fine, someone already sent the right response
            return

        # Bad response: replace it
        options.remove(existing)

    # Add our option
    options.append(new_status_code)



"""
Option handlers for the basic :rfc:`3315` options
"""

import logging

from dhcpkit.ipv6.option_handlers import CopyRelayOptionHandler
from dhcpkit.ipv6.options import InterfaceIdOption

logger = logging.getLogger(__name__)


class InterfaceIdOptionHandler(CopyRelayOptionHandler):
    """
    The handler for InterfaceIdOptions in relay messages
    """

    def __init__(self):
        super().__init__(InterfaceIdOption)



"""
Test the RelayServerMessage implementation
"""
from ipaddress import IPv6Address
import unittest

from dhcpkit.ipv6.messages import RelayServerMessage, RelayForwardMessage, UnknownMessage, Message
from dhcpkit.ipv6.options import RelayMessageOption
from tests.ipv6.messages import test_message


class RelayServerMessageTestCase(test_message.MessageTestCase):
    def setUp(self):
        # The following attributes must be overruled by child classes
        # The basics are tested with nested RelayForwardMessages
        self.packet_fixture = bytes.fromhex('0c'  # message_type: MSG_RELAY_FORW
                                            '02'  # hop_count
                                            '20010db8000000000000000000020001'  # link_address
                                            '20010db8000000000000000000020002'  # peer_address
                                            '0009'  # option_type: OPTION_RELAY_MSG
                                            '0051'  # option_length
                                            '0c'  # message_type: MSG_RELAY_FORW
                                            '01'  # hop_count
                                            '20010db8000000000000000000010001'  # link_address
                                            '20010db8000000000000000000010002'  # peer_address
                                            '0009'  # option_type: OPTION_RELAY_MSG
                                            '002b'  # option_length
                                            '0c'  # message_type: MSG_RELAY_FORW
                                            '00'  # hop_count
                                            '20010db8000000000000000000000001'  # link_address
                                            '20010db8000000000000000000000002'  # peer_address
                                            '0009'  # option_type: OPTION_RELAY_MSG
                                            '0005'  # option_length
                                            'ff'  # some unknown  message_type: 255
                                            '41424344')  # some random message_data: 'ABCD'

        self.message_fixture = RelayForwardMessage(
            hop_count=2,
            link_address=IPv6Address('2001:db8::2:1'),
            peer_address=IPv6Address('2001:db8::2:2'),
            options=[
                RelayMessageOption(
                    relayed_message=RelayForwardMessage(
                        hop_count=1,
                        link_address=IPv6Address('2001:db8::1:1'),
                        peer_address=IPv6Address('2001:db8::1:2'),
                        options=[
                            RelayMessageOption(
                                relayed_message=RelayForwardMessage(
                                    hop_count=0,
                                    link_address=IPv6Address('2001:db8::1'),
                                    peer_address=IPv6Address('2001:db8::2'),
                                    options=[
                                        RelayMessageOption(relayed_message=UnknownMessage(255, b'ABCD'))
                                    ]))
                        ]))
            ])

        self.parse_packet()

    def parse_packet(self):
        super().parse_packet()
        self.assertIsInstance(self.message, RelayServerMessage)

    def test_validate_hop_count(self):
        self.message.hop_count = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 8 bit integer'):
            self.message.validate()

        self.message.hop_count = 255
        self.message.validate()

        self.message.hop_count = 256
        with self.assertRaisesRegex(ValueError, 'unsigned 8 bit integer'):
            self.message.validate()

    def test_validate_link_address(self):
        self.message.link_address = bytes.fromhex('20010db8000000000000000000000001')
        with self.assertRaisesRegex(ValueError, 'Link-address .* IPv6 address'):
            self.message.validate()

        self.message.link_address = IPv6Address('ff02::1')
        with self.assertRaisesRegex(ValueError, 'Link-address .* non-multicast IPv6 address'):
            self.message.validate()

    def test_validate_peer_address(self):
        self.message.peer_address = bytes.fromhex('20010db8000000000000000000000001')
        with self.assertRaisesRegex(ValueError, 'Peer-address .* IPv6 address'):
            self.message.validate()

        self.message.peer_address = IPv6Address('ff02::1')
        with self.assertRaisesRegex(ValueError, 'Peer-address .* non-multicast IPv6 address'):
            self.message.validate()

    def test_get_relayed_message(self):
        self.assertEqual(self.message.relayed_message,
                         self.message_fixture.get_option_of_type(RelayMessageOption).relayed_message)

    def test_set_relayed_message(self):
        # Start with empty options
        self.message.options = []
        self.assertEqual(len(self.message.get_options_of_type(RelayMessageOption)), 0)

        self.message.relayed_message = UnknownMessage(255, b'ThisIsAnUnknownMessage')
        self.assertEqual(len(self.message.get_options_of_type(RelayMessageOption)), 1)
        self.assertEqual(self.message.relayed_message.message_data, b'ThisIsAnUnknownMessage')

        self.message.relayed_message = UnknownMessage(255, b'ThisIsADifferentUnknownMessage')
        self.assertEqual(len(self.message.get_options_of_type(RelayMessageOption)), 1)
        self.assertEqual(self.message.relayed_message.message_data, b'ThisIsADifferentUnknownMessage')

    def test_inner_message(self):
        # Make sure the inner message is a message that is not a RelayServerMessage
        self.assertIsInstance(self.message.inner_message, Message)
        self.assertNotIsInstance(self.message.inner_message, RelayServerMessage)

    def test_inner_relay_message(self):
        # Make sure the inner relay message is a RelayServerMessage
        self.assertIsInstance(self.message.inner_relay_message, RelayServerMessage)

        # that contains a message that is not a RelayServerMessage
        self.assertIsInstance(self.message.inner_relay_message.relayed_message, Message)
        self.assertNotIsInstance(self.message.inner_relay_message.relayed_message, RelayServerMessage)

    def test_missing_inner_message(self):
        # Remove the final message and check that it's being handled correctly
        self.message.inner_relay_message.options = []
        self.assertIsNone(self.message.inner_message)
        self.assertIsNone(self.message.inner_relay_message)


if __name__ == '__main__':
    unittest.main()



"""
Test the ReplyMessage implementation
"""
import unittest
import codecs
from ipaddress import IPv6Network, IPv6Address

from dhcpkit.ipv6.duids import LinkLayerDUID, LinkLayerTimeDUID
from dhcpkit.ipv6.messages import ReplyMessage
from dhcpkit.ipv6.options import ClientIdOption, IANAOption, \
    ReconfigureAcceptOption, IAAddressOption, ServerIdOption
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption
from dhcpkit.ipv6.extensions.dns import RecursiveNameServersOption
from tests.ipv6.messages import test_client_server_message

""

# DHCPv6
#     Message type: Reply (7)
#     Transaction ID: 0xf350d6
#     Identity Association for Non-temporary Address
#         Option: Identity Association for Non-temporary Address (3)
#         Length: 40
#         Value: c43cb2f100000000000000000005001820010db8ffff0001...
#         IAID: c43cb2f1
#         T1: 0
#         T2: 0
#         IA Address
#             Option: IA Address (5)
#             Length: 24
#             Value: 20010db8ffff0001000c00000000e09c0000017700000258
#             IPv6 address: 2001:db8:ffff:1:c::e09c (2001:db8:ffff:1:c::e09c)
#             Preferred lifetime: 375
#             Valid lifetime: 600
#     Identity Association for Prefix Delegation
#         Option: Identity Association for Prefix Delegation (25)
#         Length: 41
#         Value: c43cb2f10000000000000000001a00190000017700000258...
#         IAID: c43cb2f1
#         T1: 0
#         T2: 0
#         IA Prefix
#             Option: IA Prefix (26)
#             Length: 25
#             Value: 00000177000002583820010db8ffccfe0000000000000000...
#             Preferred lifetime: 375
#             Valid lifetime: 600
#             Prefix length: 56
#             Prefix address: 2001:db8:ffcc:fe00:: (2001:db8:ffcc:fe00::)
#     Client Identifier
#         Option: Client Identifier (1)
#         Length: 10
#         Value: 000300013431c43cb2f1
#         DUID: 000300013431c43cb2f1
#         DUID Type: link-layer address (3)
#         Hardware type: Ethernet (1)
#         Link-layer address: 34:31:c4:3c:b2:f1
#     Server Identifier
#         Option: Server Identifier (2)
#         Length: 14
#         Value: 000100011d1d49cf00137265ca42
#         DUID: 000100011d1d49cf00137265ca42
#         DUID Type: link-layer address plus time (1)
#         Hardware type: Ethernet (1)
#         DUID Time: Jun 24, 2015 12:58:23.000000000 CEST
#         Link-layer address: 00:13:72:65:ca:42
#     Reconfigure Accept
#         Option: Reconfigure Accept (20)
#         Length: 0
#     DNS recursive name server
#         Option: DNS recursive name server (23)
#         Length: 16
#         Value: 20014860486000000000000000008888
#          1 DNS server address: 2001:4860:4860::8888 (2001:4860:4860::8888)

reply_message = ReplyMessage(
    transaction_id=bytes.fromhex('f350d6'),
    options=[
        IANAOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAAddressOption(address=IPv6Address('2001:db8:ffff:1:c::e09c'), preferred_lifetime=375, valid_lifetime=600),
        ]),
        IAPDOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAPrefixOption(prefix=IPv6Network('2001:db8:ffcc:fe00::/56'), preferred_lifetime=375, valid_lifetime=600),
        ]),
        ClientIdOption(duid=LinkLayerDUID(hardware_type=1, link_layer_address=bytes.fromhex('3431c43cb2f1'))),
        ServerIdOption(duid=LinkLayerTimeDUID(hardware_type=1, time=488458703,
                                              link_layer_address=bytes.fromhex('00137265ca42'))),
        ReconfigureAcceptOption(),
        RecursiveNameServersOption(dns_servers=[IPv6Address('2001:4860:4860::8888')]),
    ],
)

reply_packet = codecs.decode('07f350d600030028c43cb2f100000000'
                             '000000000005001820010db8ffff0001'
                             '000c00000000e09c0000017700000258'
                             '00190029c43cb2f10000000000000000'
                             '001a001900000177000002583820010d'
                             'b8ffccfe000000000000000000000100'
                             '0a000300013431c43cb2f10002000e00'
                             '0100011d1d49cf00137265ca42001400'
                             '00001700102001486048600000000000'
                             '0000008888', 'hex')


class ReplyMessageTestCase(test_client_server_message.ClientServerMessageTestCase):
    def setUp(self):
        self.packet_fixture = reply_packet
        self.message_fixture = reply_message
        self.parse_packet()


if __name__ == '__main__':
    unittest.main()



"""
An option handler that assigns addresses based on DUID from a shelf file
"""
import codecs
import configparser
import logging
import shelve
from ipaddress import IPv6Network

from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.option_handlers.fixed_assignment import FixedAssignmentOptionHandler
from dhcpkit.ipv6.option_handlers.utils import Assignment
from dhcpkit.ipv6.options import ClientIdOption, InterfaceIdOption
from dhcpkit.ipv6.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


def create_shelf_from_csv():
    """
    Function to be called from the command line to convert a CSV based assignments file to a shelf.

    :return: exit code
    """
    import argparse
    import sys
    from dhcpkit.ipv6.option_handlers.csv import CSVBasedFixedAssignmentOptionHandler

    # Handle command line arguments
    parser = argparse.ArgumentParser(
        description="Assignments CSV to Shelf converter",
    )

    parser.add_argument("source", help="the source CSV file")
    parser.add_argument("destination", help="the destination shelf file")
    parser.add_argument("-v", "--verbosity", action="count", default=0, help="increase output verbosity")

    args = parser.parse_args()

    # Our logger is the root logger now
    global logger
    logger = logging.getLogger()

    # Don't filter on level in the root logger
    logger.setLevel(logging.NOTSET)

    # Output to sys.stdout
    stdout_handler = logging.StreamHandler(stream=sys.stdout)

    # Set level according to verbosity
    if args.verbosity >= 3:
        stdout_handler.setLevel(logging.DEBUG)
    elif args.verbosity == 2:
        stdout_handler.setLevel(logging.INFO)
    elif args.verbosity >= 1:
        stdout_handler.setLevel(logging.WARNING)
    else:
        stdout_handler.setLevel(logging.CRITICAL)

    logger.addHandler(stdout_handler)

    logger.info("Reading assignments from CSV file {}".format(args.source))
    assignments = CSVBasedFixedAssignmentOptionHandler.parse_csv_file(args.source)

    logger.info("Writing assignments to shelf file {}".format(args.destination))
    with shelve.open(args.destination, 'n') as shelf:
        for key, value in assignments:
            shelf[key] = value

        logger.info("Wrote {} assignments".format(len(shelf)))


class ShelfBasedFixedAssignmentOptionHandler(FixedAssignmentOptionHandler):
    """
    Assign addresses and/or prefixes based on the contents of a Shelf file
    """

    def __init__(self, filename: str, responsible_for_links: [IPv6Network],
                 address_preferred_lifetime: int, address_valid_lifetime: int,
                 prefix_preferred_lifetime: int, prefix_valid_lifetime: int):
        """
        Initialise the mapping. This handler will respond to clients on responsible_for_links and assume that all
        addresses in the mapping are appropriate for on those links.

        :param filename: The filename containing the shelf data
        :param responsible_for_links: The IPv6 links that this handler is responsible for
        """
        super().__init__(responsible_for_links,
                         address_preferred_lifetime, address_valid_lifetime,
                         prefix_preferred_lifetime, prefix_valid_lifetime)

        self.mapping = shelve.open(filename, 'r')

    def get_assignment(self, bundle: TransactionBundle) -> Assignment:
        """
        Look up the assignment based on DUID, Interface-ID of the relay closest to the client and Remote-ID of the
        relay closest to the client, in that order.

        :param bundle: The transaction bundle
        :return: The assignment, if any
        """
        # Look up based on DUID
        duid_option = bundle.request.get_option_of_type(ClientIdOption)
        duid = 'duid:' + codecs.encode(duid_option.duid.save(), 'hex').decode('ascii')
        if duid in self.mapping:
            return self.mapping[duid]

        # Look up based on Interface-ID
        interface_id_option = bundle.incoming_relay_messages[0].get_option_of_type(InterfaceIdOption)
        interface_id = None
        if interface_id_option:
            interface_id = 'interface-id:' + codecs.encode(interface_id_option.interface_id, 'hex').decode('ascii')
            if interface_id in self.mapping:
                return self.mapping[interface_id]

        # Look up based on Remote-ID
        remote_id_option = bundle.incoming_relay_messages[0].get_option_of_type(RemoteIdOption)
        remote_id = None
        if remote_id_option:
            remote_id = 'remote-id:{}:{}'.format(remote_id_option.enterprise_number,
                                                 codecs.encode(remote_id_option.remote_id, 'hex').decode('ascii'))
            if remote_id in self.mapping:
                return self.mapping[remote_id]

        # Nothing found
        identifiers = filter(bool, [duid, remote_id, interface_id])
        logger.info("No assignment found for {}".format(', '.join(identifiers)))

        return Assignment(address=None, prefix=None)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str=None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        # The option handler ID is our primary link prefix
        responsible_for_links = []
        try:
            prefix = IPv6Network(option_handler_id)
            responsible_for_links.append(prefix)
        except ValueError:
            raise configparser.ParsingError("The ID of [{}] must be the primary link prefix".format(section.name))

        # Add any extra prefixes
        additional_prefixes = section.get('additional-prefixes', '').split(' ')
        for additional_prefix in additional_prefixes:
            if not additional_prefix:
                continue

            try:
                prefix = IPv6Network(additional_prefix)
                responsible_for_links.append(prefix)
            except ValueError:
                raise configparser.ParsingError("'{}' is not a valid IPv6 prefix".format(additional_prefix))

        # Get the lifetimes
        address_preferred_lifetime = section.getint('address-preferred-lifetime', 3600)
        address_valid_lifetime = section.getint('address-valid-lifetime', 7200)
        prefix_preferred_lifetime = section.getint('prefix-preferred-lifetime', 43200)
        prefix_valid_lifetime = section.getint('prefix-valid-lifetime', 86400)

        shelf_filename = section.get('assignments-file')

        return cls(shelf_filename, responsible_for_links,
                   address_preferred_lifetime, address_valid_lifetime,
                   prefix_preferred_lifetime, prefix_valid_lifetime)



"""
Test the camelcase conversion functions
"""
import unittest

from dhcpkit.utils import camelcase_to_underscore, camelcase_to_dash


class CamelCaseTestCase(unittest.TestCase):
    def test_camelcase_to_underscore(self):
        self.assertEqual(camelcase_to_underscore('CamelCase'), 'camel_case')
        self.assertEqual(camelcase_to_underscore('CamelCASE'), 'camel_case')
        self.assertEqual(camelcase_to_underscore('CAMELCase'), 'camel_case')
        self.assertEqual(camelcase_to_underscore('MyCAMELCase'), 'my_camel_case')
        self.assertEqual(camelcase_to_underscore('Camel123Case'), 'camel123_case')
        self.assertEqual(camelcase_to_underscore('CAMEL123Case'), 'camel123_case')
        self.assertEqual(camelcase_to_underscore('Camel-Case'), 'camel_case')
        self.assertEqual(camelcase_to_underscore('camel-case'), 'camel_case')
        self.assertEqual(camelcase_to_underscore('Camel_Case'), 'camel_case')
        self.assertEqual(camelcase_to_underscore('camel_case'), 'camel_case')

    def test_camelcase_to_dash(self):
        self.assertEqual(camelcase_to_dash('CamelCase'), 'camel-case')
        self.assertEqual(camelcase_to_dash('CamelCASE'), 'camel-case')
        self.assertEqual(camelcase_to_dash('CAMELCase'), 'camel-case')
        self.assertEqual(camelcase_to_dash('MyCAMELCase'), 'my-camel-case')
        self.assertEqual(camelcase_to_dash('Camel123Case'), 'camel123-case')
        self.assertEqual(camelcase_to_dash('CAMEL123Case'), 'camel123-case')
        self.assertEqual(camelcase_to_dash('Camel-Case'), 'camel-case')
        self.assertEqual(camelcase_to_dash('camel-case'), 'camel-case')
        self.assertEqual(camelcase_to_dash('Camel_Case'), 'camel-case')
        self.assertEqual(camelcase_to_dash('camel_case'), 'camel-case')

if __name__ == '__main__':
    unittest.main()



"""
Test the IPv6 utility functions
"""
from ipaddress import IPv6Network, IPv6Address
import unittest

from dhcpkit.ipv6.utils import address_in_prefixes, prefix_overlaps_prefixes


class IPv6UtilsTestCase(unittest.TestCase):
    def test_address_in_prefixes(self):
        prefixes = [IPv6Network('2001:db8::/48'), IPv6Network('2001:db8:1:2::/64')]
        good_address = IPv6Address('2001:db8::1')
        bad_address = IPv6Address('2001:db8:1::1')

        self.assertTrue(address_in_prefixes(good_address, prefixes))
        self.assertFalse(address_in_prefixes(bad_address, prefixes))

    def test_prefix_overlaps_prefixes(self):
        prefixes = [IPv6Network('2001:db8::/48'), IPv6Network('2001:db8:1:2::/64')]
        good_prefix = IPv6Network('2001:db8::/64')
        bad_prefix = IPv6Network('2001:db8:1::/64')

        self.assertTrue(prefix_overlaps_prefixes(good_prefix, prefixes))
        self.assertFalse(prefix_overlaps_prefixes(bad_prefix, prefixes))


if __name__ == '__main__':
    unittest.main()



"""
Base classes for DHCP message handlers
"""

from abc import ABC, abstractmethod
import configparser
import logging

from dhcpkit.ipv6.messages import RelayServerMessage, Message
from dhcpkit.rwlock import RWLock

logger = logging.getLogger(__name__)


class MessageHandler(ABC):
    """
    This is the base class for all message handlers. Subclassing this provides the most flexibility but it does require
    the most effort to implement correctly as well.
    """

    def __init__(self, config: configparser.ConfigParser):
        """
        Initialise the handler. The config is provided from the configuration file, which is guaranteed to have a
        [handler] section.

        Objects of this class *MUST* be thread-safe.

        :param config: Contents of the configuration file
        """
        self.config = config

        # Provide a lock so that our state can be protected during a reload
        self.lock = RWLock()

        # Implement initialisation as a reload
        self.handle_reload()

    def reload(self, new_config: configparser.ConfigParser):
        """
        This is called by the server on SIGHUP so the configuration can be reloaded, caches can be cleared etc.

        Subclasses shouldn't overwrite this method but the handle_reload() method, which will automatically be
        protected with a lock.

        :param new_config: The new configuration after the reload
        """
        with self.lock.write_lock():
            self.config = new_config
            self.handle_reload()

    # noinspection PyMethodMayBeStatic
    def handle_reload(self):
        """
        This method can be overwritten by subclasses to handle configuration reloads.
        """
        pass

    @abstractmethod
    def handle(self, received_message: RelayServerMessage, received_over_multicast: bool) -> Message or None:
        """
        The main dispatcher for incoming messages. Subclasses must overwrite this. The incoming message is always
        wrapped in a RelayForwardMessage. That is: our server acts like an 'internal' relay. This way the interface
        information is captured for processing without having to differentiate between relayed and non-relayed
        messages in the handling logic.

        :param received_message: The parsed incoming request, wrapped in an 'internal' RelayServerMessage
        :param received_over_multicast: Whether the request was received over multicast
        :returns: The message to reply with
        """



"""
A base handler that decodes the incoming request and dispatches it to methods that implement standard DHCP server
behaviour.
"""

import configparser
import logging

from dhcpkit.ipv6.duids import DUID
from dhcpkit.ipv6.exceptions import CannotRespondError, UseMulticastError
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption
from dhcpkit.ipv6.message_handlers import MessageHandler
from dhcpkit.ipv6.messages import ClientServerMessage, ReplyMessage, AdvertiseMessage
from dhcpkit.ipv6.messages import Message, RelayServerMessage, SolicitMessage, RequestMessage, ConfirmMessage, \
    RenewMessage, RebindMessage, InformationRequestMessage, ReleaseMessage, DeclineMessage
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.option_handlers.basic import ClientIdOptionHandler, ServerIdOptionHandler, \
    ConfirmStatusOptionHandler, ReleaseStatusOptionHandler, DeclineStatusOptionHandler
from dhcpkit.ipv6.option_handlers.interface_id import InterfaceIdOptionHandler
from dhcpkit.ipv6.option_handlers.rapid_commit import RapidCommitOptionHandler
from dhcpkit.ipv6.option_handlers.unanswered import UnansweredIAPDOptionHandler, UnansweredIAOptionHandler
from dhcpkit.ipv6.options import ClientIdOption, ServerIdOption, StatusCodeOption, STATUS_USEMULTICAST, \
    IAAddressOption, IANAOption, IATAOption
from dhcpkit.ipv6.transaction_bundle import TransactionBundle
from dhcpkit.utils import camelcase_to_underscore

logger = logging.getLogger(__name__)


class StandardMessageHandler(MessageHandler):
    """
    This is the base class for standard handlers. It implements the standard handling of the DHCP protocol. Subclasses
    only need to provide the right addresses and options.

    :type server_duid: DUID
    :type allow_rapid_commit: bool
    :type rapid_commit_rejections: bool
    :type option_handlers: list[OptionHandler]
    """

    server_duid = None
    allow_rapid_commit = False
    rapid_commit_rejections = False
    option_handlers = None

    def handle_reload(self):
        """
        Reconstruct the DUID and all option handlers from the data in the configuration.
        """
        from dhcpkit.ipv6.option_handler_registry import option_handler_registry

        # Parse this once so we don't have to re-parse at every request
        duid_bytes = bytes.fromhex(self.config['server']['duid'])
        length, self.server_duid = DUID.parse(duid_bytes, length=len(duid_bytes))

        # Allow rapid commit?
        self.allow_rapid_commit = self.config.getboolean('server', 'allow-rapid-commit', fallback=False)
        self.rapid_commit_rejections = self.config.getboolean('server', 'rapid-commit-rejections', fallback=False)

        # Build the option handlers
        self.option_handlers = []

        if self.allow_rapid_commit:
            # Rapid commit happens as the first thing in the post() stage
            self.option_handlers.append(RapidCommitOptionHandler(self.rapid_commit_rejections))

        # These are mandatory
        self.option_handlers.append(ServerIdOptionHandler(duid=self.server_duid))
        self.option_handlers.append(ClientIdOptionHandler())
        self.option_handlers.append(InterfaceIdOptionHandler())

        # Add the ones from the configuration
        for section_name in self.config.sections():
            parts = section_name.split(' ')
            if parts[0] != 'option':
                # Not an option
                continue

            option_handler_name = parts[1]
            option_handler_id = len(parts) > 2 and parts[2] or None
            option_handler_class = option_handler_registry.get(option_handler_name)
            if not option_handler_class or not issubclass(option_handler_class, OptionHandler):
                raise configparser.ParsingError("Unknown option handler: {}".format(option_handler_name))

            logger.debug("Creating {} from config".format(option_handler_class.__name__))
            option = option_handler_class.from_config(self.config[section_name], option_handler_id=option_handler_id)
            self.option_handlers.append(option)

        # Add cleanup handlers so they run last in the handling phase
        self.option_handlers.append(UnansweredIAOptionHandler())
        self.option_handlers.append(UnansweredIAPDOptionHandler())

        # Confirm/Release/Decline messages always need a status
        self.option_handlers.append(ConfirmStatusOptionHandler())
        self.option_handlers.append(ReleaseStatusOptionHandler())
        self.option_handlers.append(DeclineStatusOptionHandler())

    @staticmethod
    def determine_method_name(request: ClientServerMessage) -> str:
        """
        Automatically determine the method name that can handle this type of message. The default implementation
        bases this on the name of the class of the message, but subclasses may provide a different behaviour.

        :param request: The incoming ClientServer Message
        :return: The name of the method that can handle this request
        """
        class_name = request.__class__.__name__
        underscored = camelcase_to_underscore(class_name)
        return 'handle_' + underscored

    def construct_use_multicast_reply(self, bundle: TransactionBundle):
        """
        Construct a message signalling to the client that they should have used multicast.

        :param bundle: The transaction bundle containing the incoming request
        :return: The proper answer to tell a client to use multicast
        """
        # Make sure we only tell this to requests that came in over multicast
        if not bundle.received_over_multicast:
            return None

        return ReplyMessage(bundle.request.transaction_id, options=[
            bundle.request.get_option_of_type(ClientIdOption),
            ServerIdOption(duid=self.server_duid),
            StatusCodeOption(STATUS_USEMULTICAST, "You cannot send requests directly to this server, "
                                                  "please use the proper multicast addresses")
        ])

    @staticmethod
    def init_response(bundle: TransactionBundle):
        """
        Create the message object in bundle.response

        :param bundle: The transaction bundle
        """
        # Start building the response
        if isinstance(bundle.request, SolicitMessage):
            bundle.response = AdvertiseMessage(bundle.request.transaction_id)

        elif isinstance(bundle.request, (RequestMessage, RenewMessage, RebindMessage,
                                         ReleaseMessage, DeclineMessage, InformationRequestMessage)):
            bundle.response = ReplyMessage(bundle.request.transaction_id)

        elif isinstance(bundle.request, ConfirmMessage):
            # Receipt of Confirm Messages: If [...] there were no addresses in any of the IAs sent by the client, the
            # server MUST NOT send a reply to the client.
            found = False
            for option in bundle.request.get_options_of_type((IANAOption, IATAOption, IAPDOption)):
                if option.get_options_of_type((IAAddressOption, IAPrefixOption)):
                    # Found an address or prefix option
                    found = True
                    break

            if not found:
                raise CannotRespondError

            bundle.response = ReplyMessage(bundle.request.transaction_id)

        else:
            logger.warning("Do not know how to reply to {}".format(type(bundle.request).__name__))
            raise CannotRespondError

        # Build the plain chain of relay reply messages
        bundle.create_outgoing_relay_messages()

    def handle(self, received_message: RelayServerMessage, received_over_multicast: bool) -> Message or None:
        """
        The main dispatcher for incoming messages.

        :param received_message: The parsed incoming request
        :param received_over_multicast: Whether the request was received over multicast
        :returns: The message to reply with
        """

        bundle = TransactionBundle(incoming_message=received_message,
                                   received_over_multicast=received_over_multicast,
                                   allow_rapid_commit=self.allow_rapid_commit)

        if not bundle.request:
            # Nothing to do...
            return None

        # Lock and handle
        with self.lock.read_lock():
            try:
                # Pre-process the request
                for option_handler in self.option_handlers:
                    option_handler.pre(bundle)

                # Init the response
                self.init_response(bundle)

                # Process the request
                for option_handler in self.option_handlers:
                    option_handler.handle(bundle)

                # Post-process the request
                for option_handler in self.option_handlers:
                    option_handler.post(bundle)
            except CannotRespondError:
                bundle.response = None
            except UseMulticastError:
                bundle.response = self.construct_use_multicast_reply(bundle)

        return bundle.outgoing_message


handler = StandardMessageHandler



"""
An option handler that assigns addresses based on DUID from a CSV file
"""
import codecs
import configparser
import csv
import logging
from ipaddress import IPv6Address, IPv6Network

from dhcpkit.ipv6.duids import DUID
from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.option_handlers.fixed_assignment import FixedAssignmentOptionHandler
from dhcpkit.ipv6.option_handlers.utils import Assignment
from dhcpkit.ipv6.options import ClientIdOption, InterfaceIdOption
from dhcpkit.ipv6.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class CSVBasedFixedAssignmentOptionHandler(FixedAssignmentOptionHandler):
    """
    Assign addresses and/or prefixes based on the contents of a CSV file
    """

    def __init__(self, filename: str, responsible_for_links: [IPv6Network],
                 address_preferred_lifetime: int, address_valid_lifetime: int,
                 prefix_preferred_lifetime: int, prefix_valid_lifetime: int):
        """
        Initialise the mapping. This handler will respond to clients on responsible_for_links and assume that all
        addresses in the mapping are appropriate for on those links.

        :param filename: The filename containing the CSV data
        :param responsible_for_links: The IPv6 links that this handler is responsible for
        """
        super().__init__(responsible_for_links,
                         address_preferred_lifetime, address_valid_lifetime,
                         prefix_preferred_lifetime, prefix_valid_lifetime)

        self.mapping = self.read_csv_file(filename)

    def get_assignment(self, bundle: TransactionBundle) -> Assignment:
        """
        Look up the assignment based on DUID, Interface-ID of the relay closest to the client and Remote-ID of the
        relay closest to the client, in that order.

        :param bundle: The transaction bundle
        :return: The assignment, if any
        """
        # Look up based on DUID
        duid_option = bundle.request.get_option_of_type(ClientIdOption)
        duid = 'duid:' + codecs.encode(duid_option.duid.save(), 'hex').decode('ascii')
        if duid in self.mapping:
            return self.mapping[duid]

        # Look up based on Interface-ID
        interface_id_option = bundle.incoming_relay_messages[0].get_option_of_type(InterfaceIdOption)
        interface_id = None
        if interface_id_option:
            interface_id = 'interface-id:' + codecs.encode(interface_id_option.interface_id, 'hex').decode('ascii')
            if interface_id in self.mapping:
                return self.mapping[interface_id]

        # Look up based on Remote-ID
        remote_id_option = bundle.incoming_relay_messages[0].get_option_of_type(RemoteIdOption)
        remote_id = None
        if remote_id_option:
            remote_id = 'remote-id:{}:{}'.format(remote_id_option.enterprise_number,
                                                 codecs.encode(remote_id_option.remote_id, 'hex').decode('ascii'))
            if remote_id in self.mapping:
                return self.mapping[remote_id]

        # Nothing found
        identifiers = filter(bool, [duid, remote_id, interface_id])
        logger.info("No assignment found for {}".format(', '.join(identifiers)))

        return Assignment(address=None, prefix=None)

    def read_csv_file(self, csv_filename: str) -> {str: Assignment}:
        """
        Read the assignments from the file specified in the configuration

        :param csv_filename: The filename of the CSV file
        :return: A dictionary mapping identifiers to assignments
        """
        assignments = dict(self.parse_csv_file(csv_filename))
        logger.info("Loaded {} assignments from CSV".format(len(assignments)))
        return assignments

    @staticmethod
    def parse_csv_file(csv_filename: str) -> [(str, Assignment)]:
        """
        Read the assignments from the file specified in the configuration

        :param csv_filename: The filename of the CSV file
        :return: An list of identifiers and their assignment
        """

        logger.debug("Loading assignments from {}".format(csv_filename))

        with open(csv_filename) as csv_file:
            # Auto-detect the CSV dialect
            sniffer = csv.Sniffer()
            sample = csv_file.read(10240)
            dialect = sniffer.sniff(sample)

            # If there is no header: assume that the columns are 'id', 'address' and 'prefix' in that order
            csv_has_header = sniffer.has_header(sample)
            fieldnames = ['id', 'address', 'prefix'] if not csv_has_header else None

            # Restart and parse
            csv_file.seek(0)
            reader = csv.DictReader(csv_file, dialect=dialect, fieldnames=fieldnames)

            # First line is column headings
            line = 1
            for row in reader:
                line += 1
                try:
                    address_str = row['address'].strip()
                    address = address_str and IPv6Address(address_str) or None

                    prefix_str = row['prefix'].strip()
                    prefix = prefix_str and IPv6Network(prefix_str) or None

                    # Validate and normalise id input
                    row_id = row['id']

                    if row_id.startswith('duid:'):
                        duid_hex = row_id.split(':', 1)[1]
                        duid_bytes = codecs.decode(duid_hex, 'hex')
                        length, duid = DUID.parse(duid_bytes, length=len(duid_bytes))
                        duid_hex = codecs.encode(duid.save(), 'hex').decode('ascii')
                        row_id = 'duid:{}'.format(duid_hex)

                    elif row_id.startswith('interface-id:'):
                        interface_id_hex = row_id.split(':', 1)[1]
                        interface_id = codecs.decode(interface_id_hex, 'hex')
                        interface_id_hex = codecs.encode(interface_id, 'hex').decode('ascii')
                        row_id = 'interface_id:{}'.format(interface_id_hex)

                    elif row_id.startswith('interface-id-str:'):
                        interface_id = row_id.split(':', 1)[1]
                        interface_id_hex = codecs.encode(interface_id.encode('ascii'), 'hex').decode('ascii')
                        row_id = 'interface_id:{}'.format(interface_id_hex)

                    elif row_id.startswith('remote-id:') or row_id.startswith('remote-id-str:'):
                        remote_id_data = row_id.split(':', 1)[1]
                        try:
                            enterprise_id, remote_id = remote_id_data.split(':', 1)
                            enterprise_id = int(enterprise_id)
                            if row_id.startswith('remote-id:'):
                                remote_id = codecs.decode(remote_id, 'hex')
                            else:
                                remote_id = remote_id.encode('ascii')

                            row_id = 'remote-id:{}:{}'.format(enterprise_id,
                                                              codecs.encode(remote_id, 'hex').decode('ascii'))
                        except ValueError:
                            raise ValueError("Remote-ID must be formatted as 'remote-id:<enterprise>:<remote-id-hex>', "
                                             "for example: 'remote-id:9:0123456789abcdef")

                    else:
                        raise ValueError("The id must start with duid: or interface-id: followed by a hex-encoded "
                                         "value, interface-id-str: followed by an ascii string, remote-id: followed by "
                                         "an enterprise-id, a colon and a hex-encoded value or remote-id-str: followed"
                                         "by an enterprise-id, a colon and an ascii string")

                    # Store the normalised id
                    logger.debug("Loaded assignment for {}".format(row_id))
                    yield row_id, Assignment(address=address, prefix=prefix)

                except KeyError:
                    raise configparser.Error("Assignment CSV must have columns 'id', 'address' and 'prefix'")
                except ValueError as e:
                    logger.error("Ignoring line {} with invalid value: {}".format(line, e))

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        # The option handler ID is our primary link prefix
        responsible_for_links = []
        try:
            prefix = IPv6Network(option_handler_id)
            responsible_for_links.append(prefix)
        except ValueError:
            raise configparser.ParsingError("The ID of [{}] must be the primary link prefix".format(section.name))

        # Add any extra prefixes
        additional_prefixes = section.get('additional-prefixes', '').split(' ')
        for additional_prefix in additional_prefixes:
            if not additional_prefix:
                continue

            try:
                prefix = IPv6Network(additional_prefix)
                responsible_for_links.append(prefix)
            except ValueError:
                raise configparser.ParsingError("'{}' is not a valid IPv6 prefix".format(additional_prefix))

        # Get the lifetimes
        address_preferred_lifetime = section.getint('address-preferred-lifetime', 3600)
        address_valid_lifetime = section.getint('address-valid-lifetime', 7200)
        prefix_preferred_lifetime = section.getint('prefix-preferred-lifetime', 43200)
        prefix_valid_lifetime = section.getint('prefix-valid-lifetime', 86400)

        csv_filename = section.get('assignments-file')

        return cls(csv_filename, responsible_for_links,
                   address_preferred_lifetime, address_valid_lifetime,
                   prefix_preferred_lifetime, prefix_valid_lifetime)



"""
Classes and constants for the DUIDs defined in :rfc:`3315`
"""
from struct import unpack_from, pack

from dhcpkit.protocol_element import ProtocolElement

# DUID type codes

DUID_LLT = 1
DUID_EN = 2
DUID_LL = 3


# This subclass remains abstract
# noinspection PyAbstractClass
class DUID(ProtocolElement):
    """
    :rfc:`3315#section-9.1`

    A DUID consists of a two-octet type code represented in network byte
    order, followed by a variable number of octets that make up the
    actual identifier.  A DUID can be no more than 128 octets long (not
    including the type code).
    """

    # This needs to be overwritten in subclasses
    duid_type = 0

    def __hash__(self) -> int:
        """
        Make DUIDs hashable.

        :return: The hash value
        """
        return hash(self.save())

    @classmethod
    def determine_class(cls, buffer: bytes, offset: int = 0) -> type:
        """
        Return the appropriate subclass from the registry, or UnknownDUID if no subclass is registered.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :return: The best known class for this duid data
        """
        from dhcpkit.ipv6.duid_registry import duid_registry
        duid_type = unpack_from('!H', buffer, offset=offset)[0]
        return duid_registry.get(duid_type, UnknownDUID)

    def parse_duid_header(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Parse the DUID type and perform some basic validation.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        if not length:
            raise ValueError('DUIDs length must be explicitly provided when parsing')

        duid_type = unpack_from('!H', buffer, offset=offset)[0]
        my_offset = 2

        if duid_type != self.duid_type:
            raise ValueError('The provided buffer does not contain {} data'.format(self.__class__.__name__))

        return my_offset


class UnknownDUID(DUID):
    """
    Container for raw DUID content for cases where we don't know how to decode the DUID.
    """

    def __init__(self, duid_type: int = 0, duid_data: bytes = b''):
        self.duid_type = duid_type
        self.duid_data = duid_data

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        self.duid_type = unpack_from('!H', buffer, offset=offset)[0]
        my_offset = self.parse_duid_header(buffer, offset, length)

        duid_len = length - my_offset
        self.duid_data = buffer[offset + my_offset:offset + my_offset + duid_len]
        my_offset += duid_len

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        return pack('!H', self.duid_type) + self.duid_data


class LinkLayerTimeDUID(DUID):
    """
    :rfc:`3315#section-9.2`

    This type of DUID consists of a two octet type field containing the
    value 1, a two octet hardware type code, four octets containing a
    time value, followed by link-layer address of any one network
    interface that is connected to the DHCP device at the time that the
    DUID is generated.  The time value is the time that the DUID is
    generated represented in seconds since midnight (UTC), January 1,
    2000, modulo 2^32.  The hardware type MUST be a valid hardware type
    assigned by the IANA as described in :rfc:`826` [14].  Both the time and
    the hardware type are stored in network byte order.  The link-layer
    address is stored in canonical form, as described in :rfc:`2464` [2].

    The following diagram illustrates the format of a DUID-LLT::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |               1               |    hardware type (16 bits)    |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                        time (32 bits)                         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      .                                                               .
      .             link-layer address (variable length)              .
      .                                                               .
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    The choice of network interface can be completely arbitrary, as long
    as that interface provides a globally unique link-layer address for
    the link type, and the same DUID-LLT SHOULD be used in configuring
    all network interfaces connected to the device, regardless of which
    interface's link-layer address was used to generate the DUID-LLT.

    Clients and servers using this type of DUID MUST store the DUID-LLT
    in stable storage, and MUST continue to use this DUID-LLT even if the
    network interface used to generate the DUID-LLT is removed.  Clients
    and servers that do not have any stable storage MUST NOT use this
    type of DUID.

    Clients and servers that use this DUID SHOULD attempt to configure
    the time prior to generating the DUID, if that is possible, and MUST
    use some sort of time source (for example, a real-time clock) in
    generating the DUID, even if that time source could not be configured
    prior to generating the DUID.  The use of a time source makes it
    unlikely that two identical DUID-LLTs will be generated if the
    network interface is removed from the client and another client then
    uses the same network interface to generate a DUID-LLT.  A collision
    between two DUID-LLTs is very unlikely even if the clocks have not
    been configured prior to generating the DUID.

    This method of DUID generation is recommended for all general purpose
    computing devices such as desktop computers and laptop computers, and
    also for devices such as printers, routers, and so on, that contain
    some form of writable non-volatile storage.

    Despite our best efforts, it is possible that this algorithm for
    generating a DUID could result in a client identifier collision.  A
    DHCP client that generates a DUID-LLT using this mechanism MUST
    provide an administrative interface that replaces the existing DUID
    with a newly-generated DUID-LLT.
    """

    duid_type = DUID_LLT

    def __init__(self, hardware_type: int = 0, time: int = 0, link_layer_address: bytes = b''):
        self.hardware_type = hardware_type
        self.time = time
        self.link_layer_address = link_layer_address

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.hardware_type, int) or not (0 <= self.hardware_type < 2 ** 16):
            raise ValueError("Hardware type must be an unsigned 16 bit integer")

        if not isinstance(self.time, int) or not (0 <= self.time < 2 ** 32):
            raise ValueError("Time must be an unsigned 32 bit integer")

        if not isinstance(self.link_layer_address, bytes):
            raise ValueError("Link-layer address must be a sequence of bytes")

        if len(self.link_layer_address) > 122:
            raise ValueError("DUID-LLT link-layer address cannot be longer than 122 bytes")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = self.parse_duid_header(buffer, offset, length)

        self.hardware_type, self.time = unpack_from('!HI', buffer, offset=offset + my_offset)
        my_offset += 6

        ll_len = length - my_offset
        self.link_layer_address = buffer[offset + my_offset:offset + my_offset + ll_len]
        my_offset += ll_len

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        return pack('!HHI', self.duid_type, self.hardware_type, self.time) + self.link_layer_address


class EnterpriseDUID(DUID):
    """
    :rfc:`3315#section-9.3`

    This form of DUID is assigned by the vendor to the device.  It
    consists of the vendor's registered Private Enterprise Number as
    maintained by IANA [6] followed by a unique identifier assigned by
    the vendor.  The following diagram summarizes the structure of a
    DUID-EN::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |               2               |       enterprise-number       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |   enterprise-number (contd)   |                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
      .                           identifier                          .
      .                       (variable length)                       .
      .                                                               .
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    The source of the identifier is left up to the vendor defining it,
    but each identifier part of each DUID-EN MUST be unique to the device
    that is using it, and MUST be assigned to the device at the time it
    is manufactured and stored in some form of non-volatile storage.  The
    generated DUID SHOULD be recorded in non-erasable storage.  The
    enterprise-number is the vendor's registered Private Enterprise
    Number as maintained by IANA [6].  The enterprise-number is stored as
    an unsigned 32 bit number.

    An example DUID of this type might look like this::

      +---+---+---+---+---+---+---+---+
      | 0 | 2 | 0 | 0 | 0 |  9| 12|192|
      +---+---+---+---+---+---+---+---+
      |132|221| 3 | 0 | 9 | 18|
      +---+---+---+---+---+---+

    This example includes the two-octet type of 2, the Enterprise Number
    (9), followed by eight octets of identifier data
    (0x0CC084D303000912).
    """

    duid_type = DUID_EN

    def __init__(self, enterprise_number: int = 0, identifier: bytes = b''):
        self.enterprise_number = enterprise_number
        self.identifier = identifier

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.enterprise_number, int) or not (0 <= self.enterprise_number < 2 ** 32):
            raise ValueError("Enterprise number must be an unsigned 32 bit integer")

        if not isinstance(self.identifier, bytes):
            raise ValueError("Identifier must be a sequence of bytes")

        if len(self.identifier) > 124:
            raise ValueError("DUID-EN identifier cannot be longer than 124 bytes")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = self.parse_duid_header(buffer, offset, length)

        self.enterprise_number = unpack_from('!I', buffer, offset=offset + my_offset)[0]
        my_offset += 4

        identifier_len = length - my_offset
        self.identifier = buffer[offset + my_offset:offset + my_offset + identifier_len]
        my_offset += identifier_len

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        return pack('!HI', self.duid_type, self.enterprise_number) + self.identifier


class LinkLayerDUID(DUID):
    """
    :rfc:`3315#section-9.4`

    This type of DUID consists of two octets containing the DUID type 3,
    a two octet network hardware type code, followed by the link-layer
    address of any one network interface that is permanently connected to
    the client or server device.  For example, a host that has a network
    interface implemented in a chip that is unlikely to be removed and

    used elsewhere could use a DUID-LL.  The hardware type MUST be a
    valid hardware type assigned by the IANA, as described in :rfc:`826`
    [14].  The hardware type is stored in network byte order.  The
    link-layer address is stored in canonical form, as described in
    :rfc:`2464` [2].  The following diagram illustrates the format of a
    DUID-LL::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |               3               |    hardware type (16 bits)    |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      .                                                               .
      .             link-layer address (variable length)              .
      .                                                               .
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    The choice of network interface can be completely arbitrary, as long
    as that interface provides a unique link-layer address and is
    permanently attached to the device on which the DUID-LL is being
    generated.  The same DUID-LL SHOULD be used in configuring all
    network interfaces connected to the device, regardless of which
    interface's link-layer address was used to generate the DUID.

    DUID-LL is recommended for devices that have a permanently-connected
    network interface with a link-layer address, and do not have
    nonvolatile, writable stable storage.  DUID-LL MUST NOT be used by
    DHCP clients or servers that cannot tell whether or not a network
    interface is permanently attached to the device on which the DHCP
    client is running.
    """

    duid_type = DUID_LL

    def __init__(self, hardware_type: int = 0, link_layer_address: bytes = b''):
        self.hardware_type = hardware_type
        self.link_layer_address = link_layer_address

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.hardware_type, int) or not (0 <= self.hardware_type < 2 ** 16):
            raise ValueError("Hardware type must be an unsigned 16 bit integer")

        if not isinstance(self.link_layer_address, bytes):
            raise ValueError("Link-layer address must be a sequence of bytes")

        if len(self.link_layer_address) > 126:
            raise ValueError("DUID-LL link-layer address cannot be longer than 126 bytes")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = self.parse_duid_header(buffer, offset, length)

        self.hardware_type = unpack_from('!H', buffer, offset=offset + my_offset)[0]
        my_offset += 2

        ll_len = length - my_offset
        self.link_layer_address = buffer[offset + my_offset:offset + my_offset + ll_len]
        my_offset += ll_len

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        return pack('!HH', self.duid_type, self.hardware_type) + self.link_layer_address



"""
Test the Message implementation
"""
import unittest

from dhcpkit.ipv6.messages import Message, UnknownMessage


class MessageTestCase(unittest.TestCase):
    def setUp(self):
        # The following attributes must be overruled by child classes
        # The basics are tested with a simple UnknownMessage
        self.packet_fixture = bytes.fromhex('ff') + b'ThisIsAnUnknownMessage'
        self.message_fixture = UnknownMessage(255, b'ThisIsAnUnknownMessage')
        self.parse_packet()

    def parse_packet(self):
        self.length, self.message = Message.parse(self.packet_fixture)
        self.assertIsInstance(self.message, Message)
        self.message_class = type(self.message)

    def test_length(self):
        self.assertEqual(self.length, len(self.packet_fixture))

    def test_parse(self):
        self.assertEqual(self.message, self.message_fixture)

    def test_save_parsed(self):
        self.assertEqual(self.packet_fixture, self.message.save())

    def test_save_fixture(self):
        self.assertEqual(self.packet_fixture, self.message_fixture.save())

    def test_validate(self):
        # This should be ok
        self.message.validate()


if __name__ == '__main__':
    unittest.main()



"""
Test the SolicitMessage implementation
"""
import unittest
import codecs
from ipaddress import IPv6Network

from dhcpkit.ipv6.duids import LinkLayerDUID
from dhcpkit.ipv6.messages import SolicitMessage
from dhcpkit.ipv6.options import ElapsedTimeOption, ClientIdOption, RapidCommitOption, IANAOption, \
    ReconfigureAcceptOption, OptionRequestOption, OPTION_IA_NA, OPTION_VENDOR_OPTS, VendorClassOption
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption, OPTION_IA_PD
from dhcpkit.ipv6.extensions.dns import OPTION_DNS_SERVERS
from dhcpkit.ipv6.extensions.sntp import OPTION_SNTP_SERVERS
from dhcpkit.ipv6.extensions.ntp import OPTION_NTP_SERVER
from dhcpkit.ipv6.extensions.sol_max_rt import OPTION_SOL_MAX_RT, OPTION_INF_MAX_RT
from tests.ipv6.messages import test_client_server_message

""

# DHCPv6
#   Message type: Solicit (1)
#   Transaction ID: 0xf350d6
#   Elapsed time
#     Option: Elapsed time (8)
#     Length: 2
#     Value: 0000
#     Elapsed time: 0 ms
#   Client Identifier
#     Option: Client Identifier (1)
#     Length: 10
#     Value: 000300013431c43cb2f1
#     DUID: 000300013431c43cb2f1
#     DUID Type: link-layer address (3)
#     Hardware type: Ethernet (1)
#     Link-layer address: 34:31:c4:3c:b2:f1
#   Rapid Commit
#     Option: Rapid Commit (14)
#     Length: 0
#   Identity Association for Non-temporary Address
#     Option: Identity Association for Non-temporary Address (3)
#     Length: 12
#     Value: c43cb2f10000000000000000
#     IAID: c43cb2f1
#     T1: 0
#     T2: 0
#   Identity Association for Prefix Delegation
#     Option: Identity Association for Prefix Delegation (25)
#     Length: 41
#     Value: c43cb2f10000000000000000001a00190000000000000000...
#     IAID: c43cb2f1
#     T1: 0
#     T2: 0
#     IA Prefix
#       Option: IA Prefix (26)
#       Length: 25
#       Value: 000000000000000000000000000000000000000000000000...
#       Preferred lifetime: 0
#       Valid lifetime: 0
#       Prefix length: 0
#       Prefix address: :: (::)
#   Reconfigure Accept
#     Option: Reconfigure Accept (20)
#     Length: 0
#   Option Request
#     Option: Option Request (6)
#     Length: 16
#     Value: 00170038001f00190003001100520053
#     Requested Option code: DNS recursive name server (23)
#     Requested Option code: NTP Server (56)
#     Requested Option code: Simple Network Time Protocol Server (31)
#     Requested Option code: Identity Association for Prefix Delegation (25)
#     Requested Option code: Identity Association for Non-temporary Address (3)
#     Requested Option code: Vendor-specific Information (17)
#     Requested Option code: SOL_MAX_RT (82)
#     Requested Option code: INF_MAX_RT (83)
#   Vendor Class
#     Option: Vendor Class (16)
#     Length: 4
#     Value: 00000368
#     Enterprise ID: AVM GmbH (872)

solicit_message = SolicitMessage(
    transaction_id=bytes.fromhex('f350d6'),
    options=[
        ElapsedTimeOption(elapsed_time=0),
        ClientIdOption(duid=LinkLayerDUID(hardware_type=1, link_layer_address=bytes.fromhex('3431c43cb2f1'))),
        RapidCommitOption(),
        IANAOption(iaid=bytes.fromhex('c43cb2f1')),
        IAPDOption(iaid=bytes.fromhex('c43cb2f1'), options=[
            IAPrefixOption(prefix=IPv6Network('::/0')),
        ]),
        ReconfigureAcceptOption(),
        OptionRequestOption(requested_options=[
            OPTION_DNS_SERVERS,
            OPTION_NTP_SERVER,
            OPTION_SNTP_SERVERS,
            OPTION_IA_PD,
            OPTION_IA_NA,
            OPTION_VENDOR_OPTS,
            OPTION_SOL_MAX_RT,
            OPTION_INF_MAX_RT,
        ]),
        VendorClassOption(enterprise_number=872),
    ],
)

solicit_packet = codecs.decode('01f350d60008000200000001000a0003'
                               '00013431c43cb2f1000e00000003000c'
                               'c43cb2f1000000000000000000190029'
                               'c43cb2f10000000000000000001a0019'
                               '00000000000000000000000000000000'
                               '00000000000000000000140000000600'
                               '1000170038001f001900030011005200'
                               '530010000400000368', 'hex')


class SolicitMessageTestCase(test_client_server_message.ClientServerMessageTestCase):
    def setUp(self):
        self.packet_fixture = solicit_packet
        self.message_fixture = solicit_message
        self.parse_packet()


if __name__ == '__main__':
    unittest.main()



"""
The DUID registry
"""
from dhcpkit.registry import Registry


class DUIDRegistry(Registry):
    """
    Registry for DHCPKit IPv6 DUIDs
    """
    entry_point = 'dhcpkit.ipv6.duids'


# Instantiate the DUID registry
duid_registry = DUIDRegistry()



"""
Option handler for IANAOptions and IAPDOptions where addresses and prefixes are pre-assigned based on DUID
"""
from abc import ABCMeta, abstractmethod
from ipaddress import IPv6Network, IPv6Address
import logging

from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption
from dhcpkit.ipv6.messages import SolicitMessage, RequestMessage, ConfirmMessage, RenewMessage, RebindMessage, \
    ReleaseMessage, DeclineMessage
from dhcpkit.ipv6.option_handlers.utils import Assignment, force_status
from dhcpkit.ipv6.options import IANAOption, IAAddressOption, StatusCodeOption, STATUS_NOTONLINK, ClientIdOption
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.transaction_bundle import TransactionBundle
from dhcpkit.ipv6.utils import address_in_prefixes, prefix_overlaps_prefixes

logger = logging.getLogger(__name__)


class FixedAssignmentOptionHandler(OptionHandler, metaclass=ABCMeta):
    """
    An option handler that gives a fixed address and/or prefix to clients
    """

    def __init__(self, responsible_for_links: [IPv6Network],
                 address_preferred_lifetime: int, address_valid_lifetime: int,
                 prefix_preferred_lifetime: int, prefix_valid_lifetime: int):
        """
        Initialise the mapping. This handler will respond to clients on responsible_for_links and assume that all
        addresses in the mapping are appropriate for on those links.

        :param responsible_for_links: The IPv6 links that this handler is responsible for
        :param address_preferred_lifetime: The preferred lifetime in seconds for addresses
        :param address_valid_lifetime: The valid lifetime in seconds for addresses
        :param prefix_preferred_lifetime: The preferred lifetime in seconds for prefixes
        :param prefix_valid_lifetime: The valid lifetime in seconds for prefixes
        """
        self.responsible_for_links = responsible_for_links
        self.address_preferred_lifetime = address_preferred_lifetime
        self.address_valid_lifetime = address_valid_lifetime
        self.prefix_preferred_lifetime = prefix_preferred_lifetime
        self.prefix_valid_lifetime = prefix_valid_lifetime

    @abstractmethod
    def get_assignment(self, bundle: TransactionBundle) -> Assignment:
        """
        Subclasses override this method to determine the assignment for the request in the bundle. This MUST return
        an Assignment object, even if no addresses are provided in it.

        :param bundle: The transaction bundle
        :return: The assignment
        """

    @staticmethod
    def find_iana_option_for_address(options: [IANAOption], address: IPv6Address) -> IANAOption or None:
        """
        Find an IANAOption that contains the given address

        :param options: The list of options to search
        :param address: The address to look for
        :return: The matching option, if any
        """
        for option in options:
            for suboption in option.get_options_of_type(IAAddressOption):
                if suboption.address == address:
                    return option

        # Nothing found: default to first
        return options[0] if options else None

    @staticmethod
    def find_iapd_option_for_prefix(options: [IAPDOption], prefix: IPv6Network) -> IAPDOption or None:
        """
        Find an IAPDOption that contains the given prefix

        :param options: The list of options to search
        :param prefix: The prefix to look for
        :return: The matching option, if any
        """
        for option in options:
            for suboption in option.get_options_of_type(IAPrefixOption):
                if suboption.prefix == prefix:
                    return option

        # Nothing found: default to first
        return options[0] if options else None

    def handle(self, bundle: TransactionBundle):
        """
        The handling is so complex that we just delegate the implementation to separate methods.

        :param bundle: The transaction bundle
        """
        if isinstance(bundle.request, (SolicitMessage, RequestMessage)):
            self.handle_request(bundle)
        elif isinstance(bundle.request, ConfirmMessage):
            self.handle_confirm(bundle)
        elif isinstance(bundle.request, (RenewMessage, RebindMessage)):
            self.handle_renew_rebind(bundle)
        elif isinstance(bundle.request, (ReleaseMessage, DeclineMessage)):
            self.handle_release_decline(bundle)

    def handle_request(self, bundle: TransactionBundle):
        """
        Handle a client requesting addresses (also handles SolicitMessage)

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unanswered_iana_options()
        unanswered_iapd_options = bundle.get_unanswered_iapd_options()

        # Try to assign the prefix first: it's not dependent on the link
        if assignment.prefix:
            found_option = self.find_iapd_option_for_prefix(unanswered_iapd_options, assignment.prefix)
            if found_option:
                # Answer to this option
                logger.info("Assigning {} to {!r}".format(assignment.prefix,
                                                          bundle.request.get_option_of_type(ClientIdOption).duid))
                response_option = IAPDOption(found_option.iaid, options=[
                    IAPrefixOption(prefix=assignment.prefix,
                                   preferred_lifetime=self.prefix_preferred_lifetime,
                                   valid_lifetime=self.prefix_valid_lifetime)
                ])
                bundle.response.options.append(response_option)
                bundle.mark_handled(found_option)

        # Make sure we are responsible for this link
        link_address = bundle.get_link_address()
        if not address_in_prefixes(link_address, self.responsible_for_links):
            logger.debug("Not assigning to link {}: "
                         "doesn't match {}".format(link_address, ', '.join(map(str, self.responsible_for_links))))
            return

        if assignment.address:
            found_option = self.find_iana_option_for_address(unanswered_iana_options, assignment.address)
            if found_option:
                # Answer to this option
                logger.info("Assigning {} to {!r}".format(assignment.address,
                                                          bundle.request.get_option_of_type(ClientIdOption).duid))
                response_option = IANAOption(found_option.iaid, options=[
                    IAAddressOption(address=assignment.address,
                                    preferred_lifetime=self.address_preferred_lifetime,
                                    valid_lifetime=self.address_valid_lifetime)
                ])
                bundle.response.options.append(response_option)
                bundle.mark_handled(found_option)

    def handle_confirm(self, bundle: TransactionBundle):
        """
        Handle a client requesting confirmation

        :param bundle: The request bundle
        """
        # Make sure we are responsible for this link
        link_address = bundle.get_link_address()
        if not address_in_prefixes(link_address, self.responsible_for_links):
            logger.debug("Not confirming to link {}: "
                         "doesn't match {}".format(link_address, ', '.join(map(str, self.responsible_for_links))))
            return

        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unanswered_iana_options()

        # See if there are any addresses on a link that I am responsible for
        for option in unanswered_iana_options:
            for suboption in option.get_options_of_type(IAAddressOption):
                if suboption.address == assignment.address:
                    # This is the address from the assignment: it's ok
                    bundle.mark_handled(option)
                    continue

                if address_in_prefixes(suboption.address, self.responsible_for_links):
                    # Oops, an address on a link that I am responsible for, but it's the wrong one...
                    force_status(bundle.response.options,
                                 StatusCodeOption(STATUS_NOTONLINK,
                                                  "{} is not assigned to you".format(suboption.address)))
                    bundle.mark_handled(option)
                    return

    def handle_renew_rebind(self, bundle: TransactionBundle):
        """
        Handle a client renewing/rebinding addresses

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Client ID for logging
        client_id_option = bundle.request.get_option_of_type(ClientIdOption)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unanswered_iana_options()
        unanswered_iapd_options = bundle.get_unanswered_iapd_options()

        for option in unanswered_iapd_options:
            if assignment.prefix and prefix_overlaps_prefixes(assignment.prefix, option.get_prefixes()):
                # Overlap with our assigned prefix: take responsibility
                response_suboptions = []
                for suboption in option.get_options_of_type(IAPrefixOption):
                    if suboption.prefix == assignment.prefix:
                        # This is the correct option, renew it
                        logger.info("Renewing {} for {!r}".format(assignment.prefix, client_id_option.duid))
                        response_suboptions.append(IAPrefixOption(prefix=assignment.prefix,
                                                                  preferred_lifetime=self.address_preferred_lifetime,
                                                                  valid_lifetime=self.address_valid_lifetime))
                    else:
                        # This isn't right
                        logger.info("Withdrawing {} from {!r}".format(suboption.prefix, client_id_option.duid))
                        response_suboptions.append(IAPrefixOption(prefix=suboption.prefix,
                                                                  preferred_lifetime=0, valid_lifetime=0))

                response_option = IAPDOption(option.iaid, options=response_suboptions)
                bundle.response.options.append(response_option)
                bundle.mark_handled(option)

        for option in unanswered_iana_options:
            if any([address_in_prefixes(address, self.responsible_for_links) for address in option.get_addresses()]):
                # Overlap with our addresses: take responsibility
                response_suboptions = []
                for suboption in option.get_options_of_type(IAAddressOption):
                    if suboption.address == assignment.address:
                        # This is the correct option, renew it
                        logger.info("Renewing {} for {!r}".format(assignment.address, client_id_option.duid))
                        response_suboptions.append(IAAddressOption(address=assignment.address,
                                                                   preferred_lifetime=self.address_preferred_lifetime,
                                                                   valid_lifetime=self.address_valid_lifetime))
                    else:
                        # This isn't right
                        logger.info("Withdrawing {} from {!r}".format(suboption.address, client_id_option.duid))
                        response_suboptions.append(IAAddressOption(address=suboption.address,
                                                                   preferred_lifetime=0, valid_lifetime=0))

                response_option = IANAOption(option.iaid, options=response_suboptions)
                bundle.response.options.append(response_option)
                bundle.mark_handled(option)

    def handle_release_decline(self, bundle: TransactionBundle):
        """
        Handle a client releasing or declining resources. Doesn't really need to do anything because assignments are
        fixed. Just mark the right options as handled.

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unanswered_iana_options()
        unanswered_iapd_options = bundle.get_unanswered_iapd_options()

        for option in unanswered_iapd_options:
            if assignment.prefix and prefix_overlaps_prefixes(assignment.prefix, option.get_prefixes()):
                # Overlap with our assigned prefix: take responsibility
                bundle.mark_handled(option)

        for option in unanswered_iana_options:
            if any([address_in_prefixes(address, self.responsible_for_links) for address in option.get_addresses()]):
                # Overlap with our addresses: take responsibility
                bundle.mark_handled(option)



"""
Implementation of DNS options as specified in :rfc:`3646`.
"""

from ipaddress import IPv6Address
from struct import pack

from dhcpkit.ipv6.messages import SolicitMessage, AdvertiseMessage, RequestMessage, RenewMessage, RebindMessage, \
    InformationRequestMessage, ReplyMessage
from dhcpkit.ipv6.options import Option
from dhcpkit.utils import parse_domain_list_bytes, encode_domain_list

OPTION_DNS_SERVERS = 23
OPTION_DOMAIN_LIST = 24


class RecursiveNameServersOption(Option):
    """
    :rfc:`3646#section-3`

    The DNS Recursive Name Server option provides a list of one or more
    IPv6 addresses of DNS recursive name servers to which a client's DNS
    resolver MAY send DNS queries [1].  The DNS servers are listed in the
    order of preference for use by the client resolver.

    The format of the DNS Recursive Name Server option is::

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |      OPTION_DNS_SERVERS       |         option-len            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |            DNS-recursive-name-server (IPv6 address)           |
      |                                                               |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |            DNS-recursive-name-server (IPv6 address)           |
      |                                                               |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                              ...                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_DNS_SERVERS (23).

    option-len
        Length of the list of DNS recursive name servers in octets; must be a
        multiple of 16.

    DNS-recursive-name-server
        IPv6 address of DNS recursive name server.

    :type dns_servers: list[IPv6Address]
    """

    option_type = OPTION_DNS_SERVERS

    def __init__(self, dns_servers: [IPv6Address] = None):
        self.dns_servers = dns_servers or []
        """List of IPv6 addresses of resolving DNS servers"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.dns_servers, list):
            raise ValueError("DNS servers must be a list")

        for address in self.dns_servers:
            if not isinstance(address, IPv6Address):
                raise ValueError("DNS server must be an IPv6 address")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may
        contain more data after the structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        if option_len % 16 != 0:
            raise ValueError('DNS Servers Option length must be a multiple of 16')

        # Parse the addresses
        self.dns_servers = []
        max_offset = option_len + header_offset
        while max_offset > my_offset:
            address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
            self.dns_servers.append(address)
            my_offset += 16

        if my_offset != max_offset:
            raise ValueError('Option length does not match combined length '
                             'of included addresses')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        buffer = bytearray()
        buffer.extend(pack('!HH', self.option_type, len(self.dns_servers) * 16))
        for address in self.dns_servers:
            buffer.extend(address.packed)

        return buffer


class DomainSearchListOption(Option):
    """
    :rfc:`3646#section-4`

    The Domain Search List option specifies the domain search list the
    client is to use when resolving hostnames with DNS.  This option does
    not apply to other name resolution mechanisms.

    The format of the Domain Search List option is::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |      OPTION_DOMAIN_LIST       |         option-len            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                          searchlist                           |
      |                              ...                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_DOMAIN_LIST (24).

    option-len
        Length of the 'searchlist' field in octets.

    searchlist
        The specification of the list of domain names in the Domain Search List.

    The list of domain names in the 'searchlist' MUST be encoded as
    specified in section "Representation and use of domain names" of
    :rfc:`3315`.

    :type search_list: list[str]
    """

    option_type = OPTION_DOMAIN_LIST

    def __init__(self, search_list: [str] = None):
        self.search_list = search_list or []
        """List of domain names to use as a search list"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        for domain_name in self.search_list:
            if len(domain_name) > 255:
                raise ValueError("Domain names must be 255 characters or less")

            if any([0 >= len(label) > 63 for label in domain_name.split('.')]):
                raise ValueError("Domain labels must be 1 to 63 characters long")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        # Parse the domain labels
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        parsed_len, self.search_list = parse_domain_list_bytes(buffer, offset=offset + my_offset, length=option_len)
        my_offset += parsed_len

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the included search domains')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        domain_buffer = encode_domain_list(self.search_list)

        buffer = bytearray()
        buffer.extend(pack('!HH', self.option_type, len(domain_buffer)))
        buffer.extend(domain_buffer)
        return buffer


SolicitMessage.add_may_contain(RecursiveNameServersOption, 0, 1)
AdvertiseMessage.add_may_contain(RecursiveNameServersOption, 0, 1)
RequestMessage.add_may_contain(RecursiveNameServersOption, 0, 1)
RenewMessage.add_may_contain(RecursiveNameServersOption, 0, 1)
RebindMessage.add_may_contain(RecursiveNameServersOption, 0, 1)
InformationRequestMessage.add_may_contain(RecursiveNameServersOption, 0, 1)
ReplyMessage.add_may_contain(RecursiveNameServersOption, 0, 1)

SolicitMessage.add_may_contain(DomainSearchListOption, 0, 1)
AdvertiseMessage.add_may_contain(DomainSearchListOption, 0, 1)
RequestMessage.add_may_contain(DomainSearchListOption, 0, 1)
RenewMessage.add_may_contain(DomainSearchListOption, 0, 1)
RebindMessage.add_may_contain(DomainSearchListOption, 0, 1)
InformationRequestMessage.add_may_contain(DomainSearchListOption, 0, 1)
ReplyMessage.add_may_contain(DomainSearchListOption, 0, 1)



"""
The option handler registry
"""
from dhcpkit.registry import Registry


class OptionHandlerRegistry(Registry):
    """
    Registry for DHCPKit IPv6 Option Handlers
    """
    entry_point = 'dhcpkit.ipv6.option_handlers'


# Instantiate the option handler registry
option_handler_registry = OptionHandlerRegistry()



"""
Test the PreferenceOption implementation
"""
import unittest

from dhcpkit.ipv6.options import PreferenceOption
from tests.ipv6.options import test_option


class OptionRequestOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('00070001ff')
        self.option_object = PreferenceOption(preference=255)
        self.parse_option()

    def test_validate_preference(self):
        self.option.preference = 0.1
        with self.assertRaisesRegex(ValueError, 'unsigned 8 bit integer'):
            self.option.validate()

        self.option.preference = 0
        self.option.validate()

        self.option.preference = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 8 bit integer'):
            self.option.validate()

        self.option.preference = 255
        self.option.validate()

        self.option.preference = 256
        with self.assertRaisesRegex(ValueError, 'unsigned 8 bit integer'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 1'):
            PreferenceOption.parse(bytes.fromhex('00070002ffff'))


if __name__ == '__main__':
    unittest.main()



"""
Implementation of SIP options as specified in :rfc:`3319`.
"""

from ipaddress import IPv6Address
from struct import pack

from dhcpkit.ipv6.options import Option
from dhcpkit.utils import parse_domain_list_bytes, encode_domain_list

OPTION_SIP_SERVER_D = 21
OPTION_SIP_SERVER_A = 22


class SIPServersDomainNameListOption(Option):
    """
    :rfc:`3319#section-3.1`

    The option length is followed by a sequence of labels, encoded
    according to Section 3.1 of :rfc:`1035` [5], quoted below:

      "Domain names in messages are expressed in terms of a sequence of
      labels.  Each label is represented as a one octet length field
      followed by that number of octets.  Since every domain name ends

      with the null label of the root, a domain name is terminated by a
      length byte of zero.  The high order two bits of every length
      octet must be zero, and the remaining six bits of the length field
      limit the label to 63 octets or less.  To simplify
      implementations, the total length of a domain name (i.e., label
      octets and label length octets) is restricted to 255 octets or
      less."

      :rfc:`1035` encoding was chosen to accommodate future
      internationalized domain name mechanisms.

    The option MAY contain multiple domain names, but these SHOULD refer
    to different NAPTR records, rather than different A records.  The
    client MUST try the records in the order listed, applying the
    mechanism described in Section 4.1 of :rfc:`3263` [3] for each.  The
    client only resolves the subsequent domain names if attempts to
    contact the first one failed or yielded no common transport protocols
    between client and server or denote a domain administratively
    prohibited by client policy.  Domain names MUST be listed in order of
    preference.

      Use of multiple domain names is not meant to replace NAPTR or SRV
      records, but rather to allow a single DHCP server to indicate
      outbound proxy servers operated by multiple providers.

    The DHCPv6 option has the format shown here::

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |      OPTION_SIP_SERVER_D      |         option-length         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                 SIP Server Domain Name List                   |
      |                              ...                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_SIP_SERVER_D (21).

    option-length
        Length of the 'SIP Server Domain Name List' field in octets; variable.

    SIP Server Domain Name List
        The domain names of the SIP outbound proxy servers for the client to use.  The domain names are encoded as
        specified in Section 8 ("Representation and use of domain names") of the DHCPv6 specification [1].

    :type domain_names: list[str]
    """

    option_type = OPTION_SIP_SERVER_D

    def __init__(self, domain_names: [str] = None):
        self.domain_names = domain_names or []
        """List of domain names of SIP servers"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        for domain_name in self.domain_names:
            if len(domain_name) > 255:
                raise ValueError("Domain names must be 255 characters or less")

            if any([0 >= len(label) > 63 for label in domain_name.split('.')]):
                raise ValueError("Domain labels must be 1 to 63 characters long")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        # Parse the domain labels
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        parsed_len, self.domain_names = parse_domain_list_bytes(buffer, offset=offset + my_offset, length=option_len)
        my_offset += parsed_len

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the included domain names')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        domain_buffer = encode_domain_list(self.domain_names)

        buffer = bytearray()
        buffer.extend(pack('!HH', self.option_type, len(domain_buffer)))
        buffer.extend(domain_buffer)
        return buffer


class SIPServersAddressListOption(Option):
    """
    :rfc:`3319#section-3.2`

    This option specifies a list of IPv6 addresses indicating SIP
    outbound proxy servers available to the client.  Servers MUST be
    listed in order of preference. ::

      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |      OPTION_SIP_SERVER_A      |           option-len          |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                   SIP server (IP address)                     |
      |                                                               |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                   SIP server (IP address)                     |
      |                                                               |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                              ...                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_SIP_SERVER_A (22).

    option-length
        Length of the 'options' field in octets; must be a multiple of 16.

    SIP server
        IPv6 address of a SIP server for the client to use. The servers are listed in the order of preference for use
        by the client.
    """

    option_type = OPTION_SIP_SERVER_A

    def __init__(self, sip_servers: [IPv6Address] = None):
        self.sip_servers = sip_servers or []
        """List of IPv6 addresses of SIP servers"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.sip_servers, list):
            raise ValueError("SIP servers must be a list")

        for address in self.sip_servers:
            if not isinstance(address, IPv6Address) or \
                    address.is_link_local or \
                    address.is_loopback or \
                    address.is_multicast or \
                    address.is_unspecified:
                raise ValueError("SIP servers must be a list of routable IPv6 addresses")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        if option_len % 16 != 0:
            raise ValueError('SIP Servers Option length must be a multiple of 16')

        # Parse the addresses
        self.sip_servers = []
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        while max_offset > my_offset:
            address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
            self.sip_servers.append(address)
            my_offset += 16

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the included addresses')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        buffer = bytearray()
        buffer.extend(pack('!HH', self.option_type, len(self.sip_servers) * 16))
        for address in self.sip_servers:
            buffer.extend(address.packed)

        return buffer



"""
Test the ServerIdOption implementation
"""
import unittest

from dhcpkit.ipv6.duids import EnterpriseDUID
from dhcpkit.ipv6.options import ServerIdOption
from tests.ipv6.options import test_option


class ServerIdOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = b'\x00\x02\x00\x15\x00\x02\x00\x00\x9d\x100123456789abcde'
        self.option_object = ServerIdOption(EnterpriseDUID(40208, b'0123456789abcde'))
        self.parse_option()

    def test_validate_duid(self):
        self.option.duid = b'0123456789abcdef'
        with self.assertRaisesRegex(ValueError, 'DUID object'):
            self.option.validate()


if __name__ == '__main__':
    unittest.main()



"""
Test the RWLock implementation
"""
import unittest

from dhcpkit.rwlock import RWLock


class RWLockTestCase(unittest.TestCase):
    def setUp(self):
        self.lock = RWLock()

    def test_readers(self):
        with self.lock.read_lock():
            self.assertEqual(self.lock.readers, 1)
            self.assertEqual(self.lock.writers, 0)
            self.assertFalse(self.lock.blocked_for_readers)
            self.assertTrue(self.lock.blocked_for_writers)

            # Multiple readers can hold the lock simultaneously
            with self.lock.read_lock():
                self.assertEqual(self.lock.readers, 2)
                self.assertEqual(self.lock.writers, 0)
                self.assertFalse(self.lock.blocked_for_readers)
                self.assertTrue(self.lock.blocked_for_writers)

    def test_writers(self):
        with self.lock.write_lock():
            self.assertEqual(self.lock.readers, 0)
            self.assertEqual(self.lock.writers, 1)
            self.assertTrue(self.lock.blocked_for_readers)
            self.assertTrue(self.lock.blocked_for_writers)

    def test_unlocked(self):
        self.assertEqual(self.lock.readers, 0)
        self.assertEqual(self.lock.writers, 0)
        self.assertFalse(self.lock.blocked_for_readers)
        self.assertFalse(self.lock.blocked_for_writers)


if __name__ == '__main__':
    unittest.main()



"""
Test the encoding and parsing of domain names
"""
import unittest

from dhcpkit.utils import parse_domain_bytes, encode_domain, parse_domain_list_bytes, encode_domain_list


class DomainNameTestCase(unittest.TestCase):
    def setUp(self):
        self.good_domain_bytes = b'\x0510-ww\x08steffann\x02nl\x00'
        self.good_relative_domain_bytes = b'\x0510-ww\x08steffann\x02nl'
        self.good_domain_name = '10-ww.steffann.nl'

        self.oversized_label_bytes = b'\x0410ww\x47steffann-steffann-steffann-steffann-' \
                                     b'steffann-steffann-steffann-steffann\x02nl\x00'
        self.oversized_label_name = '10ww.steffann-steffann-steffann-steffann-steffann-steffann-steffann-steffann.nl'

        self.buffer_overflow_bytes = b'\x0410ww\x10END'
        self.unending_bytes = b'\x0410ww\x03END'

    def test_parse_good(self):
        offset, domain_name = parse_domain_bytes(self.good_domain_bytes)
        self.assertEqual(offset, len(self.good_domain_bytes))
        self.assertEqual(domain_name, self.good_domain_name)

    def test_parse_relative(self):
        offset, domain_name = parse_domain_bytes(self.good_relative_domain_bytes, allow_relative=True)
        self.assertEqual(offset, len(self.good_relative_domain_bytes))
        self.assertEqual(domain_name, self.good_domain_name)

    def test_encode_good(self):
        domain_bytes = encode_domain(self.good_domain_name)
        self.assertEqual(domain_bytes, self.good_domain_bytes)

    def test_encode_relative(self):
        domain_bytes = encode_domain(self.good_domain_name, allow_relative=True)
        self.assertEqual(domain_bytes, self.good_relative_domain_bytes)

        domain_bytes = encode_domain(self.good_domain_name + '.', allow_relative=True)
        self.assertEqual(domain_bytes, self.good_domain_bytes)

    def test_parse_oversized_label(self):
        self.assertRaisesRegex(ValueError, 'label with invalid length', parse_domain_bytes, self.oversized_label_bytes)

    def test_encode_oversized_label(self):
        self.assertRaisesRegex(ValueError, 'label with invalid length', encode_domain, self.oversized_label_name)

    def test_parse_buffer_overflow(self):
        self.assertRaisesRegex(ValueError, 'exceeds available buffer', parse_domain_bytes, self.buffer_overflow_bytes)

    def test_parse_unending(self):
        self.assertRaisesRegex(ValueError, 'must end with a 0-length label', parse_domain_bytes, self.unending_bytes)


class DomainNameListTestCase(unittest.TestCase):
    def setUp(self):
        self.good_domains_bytes = b'\x06google\x03com\x00\x0410ww\x08steffann\x02nl\x00'
        self.good_domains_list = ['google.com', '10ww.steffann.nl']

    def test_parse_good(self):
        offset, domain_names = parse_domain_list_bytes(self.good_domains_bytes)
        self.assertEqual(offset, len(self.good_domains_bytes))
        self.assertListEqual(domain_names, self.good_domains_list)

    def test_encode_good(self):
        domain_bytes = encode_domain_list(self.good_domains_list)
        self.assertEqual(domain_bytes, self.good_domains_bytes)


if __name__ == '__main__':
    unittest.main()



"""
Test the IAAddressOption implementation
"""
from ipaddress import IPv6Address
import unittest

from dhcpkit.ipv6.options import IAAddressOption, StatusCodeOption, STATUS_NOTONLINK
from tests.ipv6.options import test_option


class IAAddressOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0005003520010db800010023045678900bc0cafe0001518000093a80'
                                          '000d0019000457686572652064696420796f752067657420746861743f')
        self.option_object = IAAddressOption(
            address=IPv6Address('2001:db8:1:23:456:7890:bc0:cafe'),
            preferred_lifetime=86400,
            valid_lifetime=7 * 86400,
            options=[
                StatusCodeOption(STATUS_NOTONLINK, 'Where did you get that?')
            ]
        )
        self.parse_option()

    def test_validate_address(self):
        self.option.address = '2001:db8::1'
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('::1')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('fe80::1')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('ff02::1')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

    def test_validate_preferred_lifetime(self):
        self.option.preferred_lifetime = 0.1
        with self.assertRaisesRegex(ValueError, 'Preferred lifetime .* unsigned 32 bit integer'):
            self.option.validate()

        self.option.preferred_lifetime = -1
        with self.assertRaisesRegex(ValueError, 'Preferred lifetime .* unsigned 32 bit integer'):
            self.option.validate()

        self.option.preferred_lifetime = 2 ** 32
        with self.assertRaisesRegex(ValueError, 'Preferred lifetime .* unsigned 32 bit integer'):
            self.option.validate()

    def test_validate_valid_lifetime(self):
        self.option.valid_lifetime = 0.1
        with self.assertRaisesRegex(ValueError, 'Valid lifetime .* unsigned 32 bit integer'):
            self.option.validate()

        self.option.valid_lifetime = -1
        with self.assertRaisesRegex(ValueError, 'Valid lifetime .* unsigned 32 bit integer'):
            self.option.validate()

        self.option.valid_lifetime = 2 ** 32
        with self.assertRaisesRegex(ValueError, 'Valid lifetime .* unsigned 32 bit integer'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'length does not match'):
            IAAddressOption.parse(bytes.fromhex('0005000020010db800010023045678900bc0cafe0001518000093a80'))


if __name__ == '__main__':
    unittest.main()



"""
Option handlers for the basic :rfc:`3315` options
"""
import configparser
import logging

from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.option_handlers import CopyRelayOptionHandler, OptionHandler

logger = logging.getLogger(__name__)


class RemoteIdOptionHandler(CopyRelayOptionHandler):
    """
    The handler for RemoteIdOptions in relay messages
    """

    def __init__(self):
        super().__init__(RemoteIdOption)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str=None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        # Don't look at the options, just add this option handler
        return cls()



"""
A class to keep the receiving and sending sockets together. When receiving traffic on a link-local multicast address
the reply should be sent from a link-local address on the receiving interface. This class makes it easy to keep those
together.
"""
from ipaddress import IPv6Address
import logging
import socket

from dhcpkit.ipv6 import SERVER_PORT, CLIENT_PORT
from dhcpkit.ipv6.exceptions import ListeningSocketError, InvalidPacketError
from dhcpkit.ipv6.messages import Message, RelayForwardMessage, RelayReplyMessage
from dhcpkit.ipv6.options import RelayMessageOption, InterfaceIdOption

logger = logging.getLogger(__name__)


class ListeningSocket:
    """
    A wrapper for a normal socket that bundles a socket to listen on with a (potentially different) socket
    to send replies from.

    :param interface_name: The name of the interface
    :param listen_socket: The socket we are listening on, may be a unicast or multicast socket
    :param reply_socket: The socket replies are sent from, must be a unicast socket
    :param global_address: The global address on the listening interface

    :type interface_name: str
    :type interface_index: int
    :type listen_socket: socket.socket
    :type listen_address: IPv6Address
    :type reply_socket: socket.socket
    :type reply_address: IPv6Address
    :type global_address: IPv6Address
    """

    def __init__(self, interface_name: str, listen_socket: socket.socket, reply_socket: socket.socket=None,
                 global_address: IPv6Address=None):
        self.interface_name = interface_name
        self.interface_id = interface_name.encode('utf-8')
        self.listen_socket = listen_socket
        self.reply_socket = reply_socket
        if self.reply_socket is None:
            self.reply_socket = self.listen_socket

        # Check that we have IPv6 UDP sockets
        if self.listen_socket.family != socket.AF_INET6 or self.listen_socket.proto != socket.IPPROTO_UDP \
                or self.reply_socket.family != socket.AF_INET6 or self.reply_socket.proto != socket.IPPROTO_UDP:
            raise ListeningSocketError("Listen and reply sockets have to be IPv6 UDP sockets")

        listen_sockname = self.listen_socket.getsockname()
        reply_sockname = self.reply_socket.getsockname()

        # Check that we are on the right port
        if listen_sockname[1] != SERVER_PORT or reply_sockname[1] != SERVER_PORT:
            raise ListeningSocketError("Listen and reply sockets have to be on port {}".format(SERVER_PORT))

        # Check that they are both on the same interface
        if listen_sockname[3] != reply_sockname[3]:
            raise ListeningSocketError("Listen and reply sockets have to be on same interface")

        self.interface_index = listen_sockname[3]
        self.listen_address = IPv6Address(listen_sockname[0].split('%')[0])
        self.reply_address = IPv6Address(reply_sockname[0].split('%')[0])

        if global_address:
            self.global_address = global_address
        elif not self.listen_address.is_link_local and not self.listen_address.is_multicast:
            self.global_address = self.listen_address
        else:
            raise ListeningSocketError("Cannot determine global address on interface {}".format(self.interface_name))

        # We only support fixed address binding
        if self.listen_address.is_unspecified or self.reply_address.is_unspecified:
            raise ListeningSocketError("This server only supports listening on explicit address, not on wildcard")

        # Multicast listeners must have link-local reply addresses
        if self.listen_address.is_multicast and not self.reply_address.is_link_local:
            raise ListeningSocketError("Multicast listening addresses need link-local reply socket")

        # Non-multicast listeners need to use a single address
        if not self.listen_address.is_multicast and self.reply_socket != self.listen_socket:
            raise ListeningSocketError("Unicast listening addresses can't use separate reply sockets")

    def recv_request(self) -> RelayForwardMessage:
        """
        Receive incoming messages

        :return: The address of the sender of the message and the received message
        """
        pkt, sender = self.listen_socket.recvfrom(65536)
        try:
            length, msg_in = Message.parse(pkt)
        except ValueError as e:
            raise InvalidPacketError(str(e), sender=sender)

        # Determine the next hop count
        if isinstance(msg_in, RelayForwardMessage):
            next_hop_count = msg_in.hop_count + 1
        else:
            next_hop_count = 0

        # Construct useful log messages
        if isinstance(msg_in, RelayForwardMessage):
            inner_relay_message = msg_in.inner_relay_message
            inner_message = inner_relay_message.relayed_message

            relay_interface_id_option = inner_relay_message.get_option_of_type(InterfaceIdOption)
            if relay_interface_id_option:
                interface_id = relay_interface_id_option.interface_id
                try:
                    interface_id = interface_id.decode('ascii')
                except ValueError:
                    pass

                interface_id_str = '{} of '.format(interface_id)
            else:
                interface_id_str = ''

            logger.debug("Received {msg_type} from {client_addr} via {interface}relay {relay_addr}".format(
                msg_type=type(inner_message).__name__,
                client_addr=inner_relay_message.peer_address,
                relay_addr=sender[0],
                interface=interface_id_str))
        else:
            logger.debug("Received {msg_type} from {client_addr}".format(
                msg_type=type(msg_in).__name__,
                client_addr=sender[0]))

        # Pretend to be an internal relay and wrap the message like a relay would
        return RelayForwardMessage(hop_count=next_hop_count,
                                   link_address=self.global_address,
                                   peer_address=IPv6Address(sender[0].split('%')[0]),
                                   options=[
                                       InterfaceIdOption(interface_id=self.interface_id),
                                       RelayMessageOption(relayed_message=msg_in)
                                   ])

    def send_reply(self, message: RelayReplyMessage) -> bool:
        """
        Send a reply using the information in the outer RelayReplyMessage

        :param message: The message to reply with
        :return: Whether sending has succeeded
        """

        # Verify that the outer relay message makes sense
        if not isinstance(message, RelayReplyMessage):
            raise ValueError("The reply has to be wrapped in a RelayReplyMessage")

        if message.link_address != self.global_address:
            raise ValueError("The relay-reply link-address does not match the relay-forward link-address")

        interface_id_option = message.get_option_of_type(InterfaceIdOption)
        if interface_id_option and interface_id_option.interface_id != self.interface_id:
            # If there is an interface-id option its contents have to match
            raise ValueError("The interface-id in the reply does not match the interface-id of the request")

        reply = message.relayed_message
        if not reply:
            raise ValueError("The RelayReplyMessage does not contain a message")

        # Down to network addresses and bytes
        port = isinstance(reply, RelayReplyMessage) and SERVER_PORT or CLIENT_PORT
        destination = (str(message.peer_address), port, 0, self.interface_index)
        data = reply.save()

        data_length = len(data)
        sent_length = self.reply_socket.sendto(data, destination)
        success = data_length == sent_length

        # Construct useful log messages
        if isinstance(reply, RelayReplyMessage):
            inner_relay_message = reply.inner_relay_message
            inner_message = inner_relay_message.relayed_message

            relay_interface_id_option = inner_relay_message.get_option_of_type(InterfaceIdOption)
            if relay_interface_id_option:
                interface_id = relay_interface_id_option.interface_id
                try:
                    interface_id = interface_id.decode('ascii')
                except ValueError:
                    pass

                interface_id_str = '{} of '.format(interface_id)
            else:
                interface_id_str = ''

            if success:
                logger.debug("Sent {msg_type} to {client_addr} via {interface}relay {relay_addr}".format(
                    msg_type=type(inner_message).__name__,
                    client_addr=inner_relay_message.peer_address,
                    relay_addr=destination[0],
                    interface=interface_id_str))
            else:
                logger.error("{msg_type} to {client_addr} via {interface}relay {relay_addr} could not be sent".format(
                    msg_type=type(inner_message).__name__,
                    client_addr=inner_relay_message.peer_address,
                    relay_addr=destination[0],
                    interface=interface_id_str))
        else:
            if success:
                logger.debug("Sent {msg_type} to {client_addr}".format(
                    msg_type=type(reply).__name__,
                    client_addr=destination[0]))
            else:
                logger.error("{msg_type} to {client_addr} could not be sent".format(
                    msg_type=type(reply).__name__,
                    client_addr=destination[0]))

        return success

    def fileno(self) -> int:
        """
        The fileno of the listening socket, so this object can be used by select()

        :return: The file descriptor
        """
        return self.listen_socket.fileno()



"""
Classes and constants for the message types defined in :rfc:`3315`
"""

from ipaddress import IPv6Address

from dhcpkit.protocol_element import ProtocolElement
from dhcpkit.utils import camelcase_to_dash

MSG_SOLICIT = 1
MSG_ADVERTISE = 2
MSG_REQUEST = 3
MSG_CONFIRM = 4
MSG_RENEW = 5
MSG_REBIND = 6
MSG_REPLY = 7
MSG_RELEASE = 8
MSG_DECLINE = 9
MSG_RECONFIGURE = 10
MSG_INFORMATION_REQUEST = 11
MSG_RELAY_FORW = 12
MSG_RELAY_REPL = 13

# The registry that keeps track of which class implements which message type
# type: {int: Option}
message_registry = {}

# The registry that keeps track of which class implements which message named type
# type: {str: Option}
message_name_registry = {}


def register_message(subclass: type):
    """
    Register a new message type in the message registry.

    :param subclass: A subclass of Message that implements the message
    """
    if not issubclass(subclass, Message):
        raise TypeError('Only Messages can be registered')

    # Store based on number
    # noinspection PyUnresolvedReferences
    message_registry[subclass.message_type] = subclass

    # Store based on name
    name = subclass.__name__
    if name.endswith('Message'):
        name = name[:-7]
    name = camelcase_to_dash(name)
    message_name_registry[name] = subclass


# This subclass remains abstract
# noinspection PyAbstractClass
class Message(ProtocolElement):
    """
    The base class for DHCP messages.

    :type message_type: int
    :type from_client_to_server: bool
    :type from_server_to_client: bool
    """
    # These needs to be overwritten in subclasses
    message_type = 0
    from_client_to_server = False
    from_server_to_client = False

    @classmethod
    def determine_class(cls, buffer: bytes, offset: int=0) -> type:
        """
        Return the appropriate subclass from the registry, or UnknownClientServerMessage if no subclass is registered.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :return: The best known class for this message data
        """
        message_type = buffer[offset]
        return message_registry.get(message_type, UnknownMessage)


class UnknownMessage(Message):
    """
    Container for raw message content for cases where we don't know how to decode the message.

    :type message_data: bytes
    """

    def __init__(self, message_type: int=0, message_data: bytes=b''):
        super().__init__()
        self.message_type = message_type
        self.message_data = message_data

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        # Check if the data is bytes
        if not isinstance(self.message_type, int) or not (0 <= self.message_type < 2 ** 8):
            raise ValueError("Message type must be an unsigned 8 bit integer")

        # Check if the data is bytes
        if not isinstance(self.message_data, bytes):
            raise ValueError("Message data must be a sequence of bytes")

    def load_from(self, buffer: bytes, offset: int=0, length: int=None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = 0

        # Message always begin with a message type
        self.message_type = buffer[offset + my_offset]
        my_offset += 1

        max_length = length or (len(buffer) - offset)
        message_data_len = max_length - my_offset
        self.message_data = buffer[offset + my_offset:offset + my_offset + message_data_len]
        my_offset += message_data_len

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        buffer = bytearray()
        buffer.append(self.message_type)
        buffer.extend(self.message_data)
        return buffer


class ClientServerMessage(Message):
    """
    :rfc:`3315#section-6`

    All DHCP messages sent between clients and servers share an identical
    fixed format header and a variable format area for options.

    All values in the message header and in options are in network byte
    order.

    Options are stored serially in the options field, with no padding
    between the options.  Options are byte-aligned but are not aligned in
    any other way such as on 2 or 4 byte boundaries.

    The following diagram illustrates the format of DHCP messages sent
    between clients and servers::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |    msg-type   |               transaction-id                  |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      .                            options                            .
      .                           (variable)                          .
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    msg-type
        Identifies the DHCP message type; the available message types are listed in section 5.3.

    transaction-id
        The transaction ID for this message exchange.

    options
        Options carried in this message; options are described in section 22.

    :type transaction_id: bytes
    :type options: list[Option]
    """

    def __init__(self, transaction_id: bytes=b'\x00\x00\x00', options: []=None):
        super().__init__()
        self.transaction_id = transaction_id
        self.options = options or []

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        # Check if the transaction is 3 bytes
        if not isinstance(self.transaction_id, bytes) or len(self.transaction_id) != 3:
            raise ValueError("Transaction-id must be 3 bytes")

        # Check if all options are allowed
        self.validate_contains(self.options)
        for option in self.options:
            option.validate()

        # Make sure that all IAIDs are unique for their type
        iaids = {}
        for option in self.options:
            iaid = getattr(option, 'iaid', None)
            if iaid:
                option_class = self.get_element_class(option)
                existing = iaids.setdefault(option_class, [])
                if iaid in existing:
                    raise ValueError("IAID {} of {} is not unique".format(iaid, option_class.__name__))
                existing.append(iaid)

    def get_options_of_type(self, klass: type) -> list:
        """
        Get all options that are subclasses of the given class.

        :param klass: The class to look for
        :returns: The list of options

        :type klass: T
        :rtype: list[T()]
        """
        return [option for option in self.options if isinstance(option, klass)]

    def get_option_of_type(self, klass: type) -> object or None:
        """
        Get the first option that is a subclass of the given class.

        :param klass: The class to look for
        :returns: The option or None

        :type klass: T
        :rtype: T() or None
        """
        for option in self.options:
            if isinstance(option, klass):
                return option

    def load_from(self, buffer: bytes, offset: int=0, length: int=None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = 0

        # These message types always begin with a message type and a transaction id
        message_type = buffer[offset + my_offset]
        my_offset += 1

        if message_type != self.message_type:
            raise ValueError('The provided buffer does not contain {} data'.format(self.__class__.__name__))

        self.transaction_id = buffer[offset + my_offset:offset + my_offset + 3]
        my_offset += 3

        # Parse the options
        from dhcpkit.ipv6.options import Option

        max_length = length or (len(buffer) - offset)
        while max_length > my_offset:
            used_buffer, option = Option.parse(buffer, offset=offset + my_offset)

            self.options.append(option)
            my_offset += used_buffer

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        buffer = bytearray()
        buffer.append(self.message_type)
        buffer.extend(self.transaction_id)
        for option in self.options:
            buffer.extend(option.save())
        return buffer


class RelayServerMessage(Message):
    """
    :rfc:`3315#section-7`

    Relay agents exchange messages with servers to relay messages between
    clients and servers that are not connected to the same link.

    All values in the message header and in options are in network byte
    order.

    Options are stored serially in the options field, with no padding
    between the options.  Options are byte-aligned but are not aligned in
    any other way such as on 2 or 4 byte boundaries.

    There are two relay agent messages, which share the following format::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |    msg-type   |   hop-count   |                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
      |                                                               |
      |                         link-address                          |
      |                                                               |
      |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
      |                               |                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
      |                                                               |
      |                         peer-address                          |
      |                                                               |
      |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
      |                               |                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
      .                                                               .
      .            options (variable number and length)   ....        .
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :type hop_count: int
    :type link_address: IPv6Address
    :type peer_address: IPv6Address
    :type options: list[Option]
    """

    def __init__(self, hop_count: int=0, link_address: IPv6Address=None, peer_address: IPv6Address=None,
                 options: []=None):
        super().__init__()
        self.hop_count = hop_count
        self.link_address = link_address
        self.peer_address = peer_address
        self.options = options or []

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        # Check hop-count
        if not isinstance(self.hop_count, int) or not (0 <= self.hop_count < 2 ** 8):
            raise ValueError("Hop-count must be an unsigned 8 bit integer")

        if not isinstance(self.link_address, IPv6Address) or self.link_address.is_multicast:
            raise ValueError("Link-address must be a non-multicast IPv6 address")

        if not isinstance(self.peer_address, IPv6Address) or self.peer_address.is_multicast:
            raise ValueError("Peer-address must be a non-multicast IPv6 address")

        # Check if all options are allowed
        self.validate_contains(self.options)
        for option in self.options:
            option.validate()

    def get_options_of_type(self, klass: type) -> list:
        """
        Get all options that are subclasses of the given class.

        :param klass: The class to look for
        :returns: The list of options

        :type klass: T
        :rtype: list[T()]
        """
        return [option for option in self.options if isinstance(option, klass)]

    def get_option_of_type(self, klass) -> object or None:
        """
        Get the first option that is a subclass of the given class.

        :param klass: The class to look for
        :type klass: type(T)
        :returns: The option or None
        :rtype: T
        """
        for option in self.options:
            if isinstance(option, klass):
                return option

    @property
    def relayed_message(self) -> Message or None:
        """
        Utility method to easily get the relayed message from the RelayMessageOption inside this RelayServerMessage.

        :return: The message, if found
        """
        from dhcpkit.ipv6.options import RelayMessageOption

        relay_message_option = self.get_option_of_type(RelayMessageOption)
        if relay_message_option:
            return relay_message_option.relayed_message

        # No embedded message found
        return None

    @relayed_message.setter
    def relayed_message(self, message):
        """
        Utility method to easily set the relayed message inside this RelayServerMessage.

        :param message: The new message
        """
        from dhcpkit.ipv6.options import RelayMessageOption

        relay_message_option = self.get_option_of_type(RelayMessageOption)
        if relay_message_option:
            # Overwrite the existing message
            relay_message_option.relayed_message = message
        else:
            # Add a new one
            relay_message_option = RelayMessageOption(relayed_message=message)
            self.options.append(relay_message_option)

    @property
    def inner_message(self) -> ClientServerMessage or None:
        """
        Utility method to easily get the innermost message from the RelayMessageOption inside this RelayServerMessage.

        :return: The message, if found
        """
        from dhcpkit.ipv6.options import RelayMessageOption

        for option in self.options:
            if isinstance(option, RelayMessageOption):
                message = option.relayed_message
                if isinstance(message, RelayServerMessage):
                    return message.inner_message
                else:
                    return message

        # No embedded message found
        return None

    @property
    def inner_relay_message(self) -> Message or None:
        """
        Utility method to easily get the innermost relay message from the RelayMessageOption inside this
        RelayServerMessage.

        :return: The message, if found
        :rtype: RelayServerMessage or None
        """
        from dhcpkit.ipv6.options import RelayMessageOption

        for option in self.options:
            if isinstance(option, RelayMessageOption):
                message = option.relayed_message
                if isinstance(message, RelayServerMessage):
                    # We contain a RelayServerMessage, so we are not the innermost: delegate
                    return message.inner_relay_message
                else:
                    # We don't contain another RelayServerMessage so we are the innermost!
                    return self

        # No embedded message found
        return None

    def load_from(self, buffer: bytes, offset: int=0, length: int=None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = 0

        # These message types always begin with a message type, a hop count, the link address and the peer address
        self.message_type = buffer[offset + my_offset]
        my_offset += 1

        self.hop_count = buffer[offset + my_offset]
        my_offset += 1

        self.link_address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
        my_offset += 16

        self.peer_address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
        my_offset += 16

        # Parse the options
        from dhcpkit.ipv6.options import Option

        max_length = length or (len(buffer) - offset)
        while max_length > my_offset:
            used_buffer, option = Option.parse(buffer, offset=offset + my_offset)
            self.options.append(option)
            my_offset += used_buffer

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        buffer = bytearray()
        buffer.append(self.message_type)
        buffer.append(self.hop_count)
        buffer.extend(self.link_address.packed)
        buffer.extend(self.peer_address.packed)
        for option in self.options:
            buffer.extend(option.save())
        return buffer


class SolicitMessage(ClientServerMessage):
    """
    SOLICIT (1)
        A client sends a Solicit message to locate servers.
    """
    message_type = MSG_SOLICIT
    from_client_to_server = True


class AdvertiseMessage(ClientServerMessage):
    """
    A server sends an Advertise message to indicate that it is available for DHCP service, in response to a
    Solicit message received from a client.
    """
    message_type = MSG_ADVERTISE
    from_server_to_client = True


class RequestMessage(ClientServerMessage):
    """
    A client sends a Request message to request configuration parameters, including IP addresses, from a
    specific server.
    """
    message_type = MSG_REQUEST
    from_client_to_server = True


class ConfirmMessage(ClientServerMessage):
    """
    A client sends a Confirm message to any available server to determine whether the addresses it was assigned
    are still appropriate to the link to which the client is connected.
    """
    message_type = MSG_CONFIRM
    from_client_to_server = True


class RenewMessage(ClientServerMessage):
    """
    A client sends a Renew message to the server that originally provided the client's addresses and configuration
    parameters to extend the lifetimes on the addresses assigned to the client and to update other configuration
    parameters.
    """
    message_type = MSG_RENEW
    from_client_to_server = True


class RebindMessage(ClientServerMessage):
    """
    A client sends a Rebind message to any available server to extend the lifetimes on the addresses assigned to
    the client and to update other configuration parameters; this message is sent after a client receives no
    response to a Renew message.
    """
    message_type = MSG_REBIND
    from_client_to_server = True


class ReplyMessage(ClientServerMessage):
    """
    A server sends a Reply message containing assigned addresses and configuration parameters in response to a
    Solicit, Request, Renew, Rebind message received from a client.  A server sends a Reply message containing
    configuration parameters in response to an Information-request message.  A server sends a Reply message in
    response to a Confirm message confirming or denying that the addresses assigned to the client are appropriate
    to the link to which the client is connected.  A server sends a Reply message to acknowledge receipt of a
    Release or Decline message.
    """
    message_type = MSG_REPLY
    from_server_to_client = True


class ReleaseMessage(ClientServerMessage):
    """
    A client sends a Release message to the server that assigned addresses to the client to indicate that the
    client will no longer use one or more of the assigned addresses.
    """
    message_type = MSG_RELEASE
    from_client_to_server = True


class DeclineMessage(ClientServerMessage):
    """
    A client sends a Decline message to a server to indicate that the client has determined that one or more
    addresses assigned by the server are already in use on the link to which the client is connected.
    """
    message_type = MSG_DECLINE
    from_client_to_server = True


class ReconfigureMessage(ClientServerMessage):
    """
    A server sends a Reconfigure message to a client to inform the client that the server has new or updated
    configuration parameters, and that the client is to initiate a Renew/Reply or Information-request/Reply
    transaction with the server in order to receive the updated information.
    """
    message_type = MSG_RECONFIGURE
    from_server_to_client = True


class InformationRequestMessage(ClientServerMessage):
    """
    A client sends an Information-request message to a server to request configuration parameters without the
    assignment of any IP addresses to the client.
    """
    message_type = MSG_INFORMATION_REQUEST
    from_client_to_server = True


class RelayForwardMessage(RelayServerMessage):
    """
    A relay agent sends a Relay-forward message to relay messages to servers, either directly or through another
    relay agent.  The received message, either a client message or a Relay-forward message from another relay
    agent, is encapsulated in an option in the Relay-forward message.
    """
    message_type = MSG_RELAY_FORW
    from_client_to_server = True

    def wrap_response(self, response: ClientServerMessage) -> Message:
        """
        The incoming message was wrapped in this RelayForwardMessage. Let this RelayForwardMessage then create a
        RelayReplyMessage with the correct options and wrap the reply .

        :param response: The response that is going to be sent to the client
        :return: The RelayReplyMessage wrapping the response
        :rtype: RelayReplyMessage
        """
        from dhcpkit.ipv6.options import RelayMessageOption

        my_response = RelayReplyMessage(self.hop_count, self.link_address, self.peer_address)

        my_relayed_message = self.relayed_message
        if isinstance(my_relayed_message, RelayForwardMessage):
            # Our relayed message is another relay message: let it create its own reply
            my_response.options.append(RelayMessageOption(
                relayed_message=my_relayed_message.wrap_response(response)
            ))
        elif isinstance(my_relayed_message, ClientServerMessage):
            # Our relayed message is a ClientServerMessage, so place the response here in the RelayReplyMessage
            my_response.options.append(RelayMessageOption(relayed_message=response))

        return my_response


class RelayReplyMessage(RelayServerMessage):
    """
    A server sends a Relay-reply message to a relay agent containing a message that the relay agent delivers to a
    client.  The Relay-reply message may be relayed by other relay agents for delivery to the destination relay
    agent.

    The server encapsulates the client message as an option in the Relay-reply message, which the relay agent
    extracts and relays to the client.
    """
    message_type = MSG_RELAY_REPL
    from_server_to_client = True


# Register the classes in this file
register_message(SolicitMessage)
register_message(AdvertiseMessage)
register_message(RequestMessage)
register_message(ConfirmMessage)
register_message(RenewMessage)
register_message(RebindMessage)
register_message(ReplyMessage)
register_message(ReleaseMessage)
register_message(DeclineMessage)
register_message(ReconfigureMessage)
register_message(InformationRequestMessage)
register_message(RelayForwardMessage)
register_message(RelayReplyMessage)
